# -*- coding: utf-8 -*-
# File: segment.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for pipeline component of table segmentation. Uses row/column detector and infers segmentations by using
ious/ioas of rows and columns.
"""


from dataclasses import dataclass
from typing import List, Literal, Optional, Sequence, Tuple, Union

import numpy as np

from ..datapoint.annotation import ImageAnnotation
from ..datapoint.box import BoundingBox, global_to_local_coords, intersection_boxes, iou
from ..datapoint.image import Image
from ..extern.base import DetectionResult
from ..mapper.maputils import MappingContextManager
from ..mapper.match import match_anns_by_intersection
from ..utils.detection_types import JsonDict
from ..utils.settings import CellType, LayoutType, ObjectTypes, Relationships, TableType
from .base import PipelineComponent
from .refine import generate_html_string
from .registry import pipeline_component_registry

__all__ = ["TableSegmentationService", "SegmentationResult", "PubtablesSegmentationService"]


@dataclass
class SegmentationResult:
    """
    Simple mutable storage for segmentation results
    """

    annotation_id: str
    row_num: int
    col_num: int
    rs: int
    cs: int


def choose_items_by_iou(
    dp: Image,
    item_proposals: List[ImageAnnotation],
    iou_threshold: float,
    above_threshold: bool = True,
    reference_item_proposals: Optional[List[ImageAnnotation]] = None,
) -> Image:
    """
    Deactivate image annotations that have ious with each other above some threshold. It will deactivate an annotation
    that has iou above some threshold with another annotation and that has a lesser score.


    :param dp: image
    :param item_proposals: Annotations to choose from. If `reference_item_proposals` is None it will compare items with
                           each other
    :param iou_threshold: iou_threshold
    :param above_threshold:
    :param reference_item_proposals: Annotations as reference. If provided, it will compare `item_proposals` with
                                     `reference_item_proposals`
    """
    item_proposals_boxes = np.array(
        [
            item.image.get_embedding(dp.image_id).to_list(mode="xyxy")
            for item in item_proposals
            if item.image is not None
        ]
    )

    triangle_ind = None
    if reference_item_proposals is not None:
        reference_item_proposals_boxes = np.array(
            [
                item.image.get_embedding(dp.image_id).to_list(mode="xyxy")
                for item in reference_item_proposals
                if item.image is not None
            ]
        )

    else:
        reference_item_proposals_boxes = item_proposals_boxes
        triangle_ind = np.triu_indices(len(item_proposals))

    iou_matrix = iou(item_proposals_boxes, reference_item_proposals_boxes)

    if triangle_ind is not None:
        iou_matrix[triangle_ind] = 0

    indices_to_deactivate = np.where((iou_matrix > iou_threshold) & (iou_matrix != 1))[0]

    if above_threshold:
        for el in indices_to_deactivate:
            item_proposals[el].deactivate()

    else:
        unique_indices = np.unique(indices_to_deactivate)
        for idx, el in enumerate(item_proposals):
            if idx not in unique_indices:
                item_proposals[idx].deactivate()

    return dp


def stretch_item_per_table(
    dp: Image,
    table: ImageAnnotation,
    row_name: str,
    col_name: str,
    remove_iou_threshold_rows: float,
    remove_iou_threshold_cols: float,
) -> Image:
    """
    Stretch rows horizontally and stretch columns vertically. Since the predictor usually does not predict a box for
    lines across the entire width of the table, lines are stretched from the left to the right edge of the table if the
    y coordinates remain the same. Columns between the top and bottom of the table can be stretched in an analogous way.

    :param dp: Image
    :param table: table image annotation
    :param row_name: item name for horizontal stretching
    :param col_name:  item name for vertical stretching
    :param remove_iou_threshold_rows: iou threshold for removing overlapping rows
    :param remove_iou_threshold_cols: iou threshold for removing overlapping columns
    :return: Image
    """

    item_ann_ids = table.get_relationship(Relationships.child)

    rows = dp.get_annotation(category_names=row_name, annotation_ids=item_ann_ids)
    if table.image is None:
        raise ValueError("table.image cannot be None")
    table_embedding_box = table.image.get_embedding(dp.image_id)

    for row in rows:
        if row.image is None:
            raise ValueError("row.image cannot be None")
        row_embedding_box = row.image.get_embedding(dp.image_id)
        row_embedding_box.ulx = table_embedding_box.ulx + 1.0
        row_embedding_box.lrx = table_embedding_box.lrx - 1.0

        # updating all bounding boxes for rows
        row.image.set_embedding(
            row.annotation_id,
            BoundingBox(
                ulx=0.0,
                uly=0.0,
                height=row_embedding_box.height,
                width=row_embedding_box.width,
                absolute_coords=row_embedding_box.absolute_coords,
            ),
        )
        local_row_box = global_to_local_coords(row_embedding_box, table_embedding_box)
        row.image.set_embedding(table.annotation_id, local_row_box)

    choose_items_by_iou(dp, rows, remove_iou_threshold_rows)

    cols = dp.get_annotation(category_names=col_name, annotation_ids=item_ann_ids)

    for col in cols:
        if col.image is None:
            raise ValueError("row.image cannot be None")
        col_embedding_box = col.image.get_embedding(dp.image_id)
        col_embedding_box.uly = table_embedding_box.uly + 1.0
        col_embedding_box.lry = table_embedding_box.lry - 1.0

        # updating all bounding boxes for cols
        col.image.set_embedding(
            col.annotation_id,
            BoundingBox(
                ulx=0.0,
                uly=0.0,
                height=col_embedding_box.height,
                width=col_embedding_box.width,
                absolute_coords=col_embedding_box.absolute_coords,
            ),
        )
        local_row_box = global_to_local_coords(col_embedding_box, table_embedding_box)
        col.image.set_embedding(table.annotation_id, local_row_box)

    choose_items_by_iou(dp, cols, remove_iou_threshold_cols)

    return dp


def _tile_by_stretching_rows_left_and_rightwise(
    dp: Image, items: List[ImageAnnotation], table: ImageAnnotation, item_name: str
) -> None:
    if table.image is None:
        raise ValueError("table.image cannot be None")
    table_embedding_box = table.image.get_embedding(dp.image_id)

    tmp_item_xy = table_embedding_box.uly + 1.0 if item_name == LayoutType.row else table_embedding_box.ulx + 1.0
    for idx, item in enumerate(items):
        with MappingContextManager(
            dp_name=dp.file_name,
            filter_level="bounding box",
            image_annotation={"category_name": item.category_name, "annotation_id": item.annotation_id},
        ):
            if item.image is None:
                raise ValueError("item.image cannot be None")
            item_embedding_box = item.image.get_embedding(dp.image_id)
            if idx != len(items) - 1:
                next_item_embedding_box = items[idx + 1].image.get_embedding(dp.image_id)  # type: ignore
                tmp_next_item_xy = (
                    (item_embedding_box.lry + next_item_embedding_box.uly) / 2
                    if item_name == LayoutType.row
                    else (item_embedding_box.lrx + next_item_embedding_box.ulx) / 2
                )
            else:
                tmp_next_item_xy = (
                    table_embedding_box.lry - 1.0 if item_name == LayoutType.row else table_embedding_box.lrx - 1.0
                )

            new_embedding_box = BoundingBox(
                ulx=item_embedding_box.ulx if item_name == LayoutType.row else tmp_item_xy,
                uly=tmp_item_xy if item_name == LayoutType.row else item_embedding_box.uly,
                lrx=item_embedding_box.lrx if item_name == LayoutType.row else tmp_next_item_xy,
                lry=tmp_next_item_xy if item_name == LayoutType.row else item_embedding_box.lry,
                absolute_coords=True,
            )
            item.image.set_embedding(dp.image_id, new_embedding_box)
            tmp_item_xy = tmp_next_item_xy


def _tile_by_stretching_rows_leftwise_column_downwise(
    dp: Image, items: List[ImageAnnotation], table: ImageAnnotation, item_name: str
) -> None:
    if table.image is None:
        raise ValueError("table.image cannot be None")
    table_embedding_box = table.image.get_embedding(dp.image_id)

    tmp_item_xy = table_embedding_box.uly + 1.0 if item_name == LayoutType.row else table_embedding_box.ulx + 1.0
    for item in items:
        with MappingContextManager(
            dp_name=dp.file_name,
            filter_level="bounding box",
            image_annotation={"category_name": item.category_name, "annotation_id": item.annotation_id},
        ):
            if item.image is None:
                raise ValueError("item.image cannot be None")
            item_embedding_box = item.image.get_embedding(dp.image_id)
            new_embedding_box = BoundingBox(
                ulx=item_embedding_box.ulx if item_name == LayoutType.row else tmp_item_xy,
                uly=tmp_item_xy if item_name == LayoutType.row else item_embedding_box.uly,
                lrx=item_embedding_box.lrx,
                lry=item_embedding_box.lry,
                absolute_coords=True,
            )

            if item == items[-1]:
                new_embedding_box = BoundingBox(
                    ulx=item_embedding_box.ulx if item_name == LayoutType.row else tmp_item_xy,
                    uly=tmp_item_xy if item_name == LayoutType.row else item_embedding_box.uly,
                    lrx=item_embedding_box.lrx if item_name == LayoutType.row else table_embedding_box.lrx - 1.0,
                    lry=table_embedding_box.lry - 1.0 if item_name == LayoutType.row else item_embedding_box.lry,
                    absolute_coords=True,
                )

            tmp_item_xy = item_embedding_box.lry if item_name == LayoutType.row else item_embedding_box.lrx
            item.image.set_embedding(dp.image_id, new_embedding_box)


def tile_tables_with_items_per_table(
    dp: Image, table: ImageAnnotation, item_name: str, stretch_rule: Literal["left", "equal"] = "left"
) -> Image:
    """
    Tiling a table with items (i.e. rows or columns). To ensure that every position in a table can be assigned to a row
    or column, rows are stretched vertically and columns horizontally. The stretching takes place according to ascending
    coordinate axes. The first item is stretched to the top or right-hand edge of the table. The next item down or to
    the right is stretched to the lower or right edge of the previous item.

    :param dp: Image
    :param table: table
    :param item_name: names.C.ROW or names.C.COL
    :param stretch_rule: Tiling can be achieved by two different stretching rules for rows and columns.
                         - 'left': The upper horizontal edge of a row will be shifted up to the lower horizontal edge
                                   of the upper neighboring row. Similarly, the left sided vertical edge of a column
                                   will be shifted towards the right sided vertical edge of the left sided neighboring
                                   column.
                         - 'equal': Upper and lower horizontal edge of rows will be shifted to the middle of the gap
                                    of two neighboring rows. Similarly, left and right sided vertical edge of a column
                                    will be shifted to the middle of the gap of two neighboring columns.
    :return: Image
    """

    item_ann_ids = table.get_relationship(Relationships.child)
    items = dp.get_annotation(category_names=item_name, annotation_ids=item_ann_ids)

    if items[0].image is not None:
        items.sort(
            key=lambda x: x.image.get_embedding(dp.image_id).cx  # type: ignore
            if item_name == LayoutType.column
            else x.image.get_embedding(dp.image_id).cy  # type: ignore
        )
    else:
        items.sort(
            key=lambda x: x.bounding_box.cx  # type: ignore
            if item_name == LayoutType.column
            else x.bounding_box.cy  # type: ignore
        )

    if stretch_rule == "left":
        _tile_by_stretching_rows_leftwise_column_downwise(dp, items, table, item_name)
    else:
        _tile_by_stretching_rows_left_and_rightwise(dp, items, table, item_name)

    return dp


def stretch_items(
    dp: Image,
    table_name: str,
    row_name: str,
    col_name: str,
    remove_iou_threshold_rows: float,
    remove_iou_threshold_cols: float,
) -> Image:
    """
    Stretch rows and columns from item detector to full table length and width. See `stretch_item_per_table`

    :param dp: Image
    :param table_name: category name for a table category ann.
    :param row_name: category name for row category ann
    :param col_name: category name for column category ann
    :param remove_iou_threshold_rows: iou threshold for removing overlapping rows
    :param remove_iou_threshold_cols: iou threshold for removing overlapping columns
    :return: An Image
    """
    table_anns = dp.get_annotation_iter(category_names=table_name)

    for table in table_anns:
        dp = stretch_item_per_table(dp, table, row_name, col_name, remove_iou_threshold_rows, remove_iou_threshold_cols)

    return dp


def _default_segment_table(cells: List[ImageAnnotation]) -> List[SegmentationResult]:
    """
    Error segmentation handling when segmentation goes wrong. It will generate a default segmentation, e.g. no real
    segmentation.

    :param cells: list of all cells of one table
    :return: list of segmentation results
    """
    raw_table_segments = []
    for cell in cells:
        raw_table_segments.append(
            SegmentationResult(annotation_id=cell.annotation_id, row_num=0, col_num=0, rs=0, cs=0)
        )
    return raw_table_segments


def segment_table(
    dp: Image,
    table: ImageAnnotation,
    item_names: Union[ObjectTypes, Sequence[ObjectTypes]],
    cell_names: Union[ObjectTypes, Sequence[ObjectTypes]],
    segment_rule: Literal["iou", "ioa"],
    threshold_rows: float,
    threshold_cols: float,
) -> List[SegmentationResult]:
    """
    Segments a table,i.e. produces for each cell a SegmentationResult. It uses numbered rows and columns that have to
    be predicted by an appropriate detector. E.g. for calculating row and rwo spans it first infers the iou of a cell
    with all rows. All ious with rows above iou_threshold_rows will induce the cell to have that row number. As there
    might be several rows, the row number of the cell will be the smallest of the number of all intersected rows. The
    row span will be equal the number of all rows with iou above the iou threshold.

    :param dp: A datapoint
    :param table: the table as image annotation.
    :param item_names: A list of item names (e.g. "row" and "column")
    :param cell_names: A list of cell names (e.g. "cell")
    :param segment_rule: 'iou' or 'ioa'
    :param threshold_rows: the iou/ioa threshold of a cell with a row in order to conclude that the cell belongs
                               to the row.
    :param threshold_cols: the iou/ioa threshold of a cell with a column in order to conclude that the cell belongs
                               to the column.
    :return: A list of len(number of cells) of SegmentationResult.
    """

    child_ann_ids = table.get_relationship(Relationships.child)
    cell_index_rows, row_index, _, _ = match_anns_by_intersection(
        dp,
        item_names[0],
        cell_names,
        segment_rule,
        threshold_rows,
        True,
        child_ann_ids,
        child_ann_ids,
    )

    cell_index_cols, col_index, _, _ = match_anns_by_intersection(
        dp,
        item_names[1],
        cell_names,
        segment_rule,
        threshold_cols,
        True,
        child_ann_ids,
        child_ann_ids,
    )

    cells = dp.get_annotation(annotation_ids=child_ann_ids, category_names=cell_names)

    rows = dp.get_annotation(annotation_ids=child_ann_ids, category_names=item_names[0])
    columns = dp.get_annotation(annotation_ids=child_ann_ids, category_names=item_names[1])

    raw_table_segments = []
    with MappingContextManager(dp_name=dp.file_name) as segment_mapping_context:
        for idx, cell in enumerate(cells):
            cell_positions_rows = cell_index_rows == idx
            rows_of_cell = [rows[k] for k in row_index[cell_positions_rows]]
            rs = np.count_nonzero(cell_index_rows == idx)
            if len(rows_of_cell):
                row_number = min([int(row.get_sub_category(CellType.row_number).category_id) for row in rows_of_cell])
            else:
                row_number = 0

            cell_positions_cols = cell_index_cols == idx
            cols_of_cell = [columns[k] for k in col_index[cell_positions_cols]]
            cs = np.count_nonzero(cell_index_cols == idx)
            if len(cols_of_cell):
                col_number = min(
                    [int(col.get_sub_category(CellType.column_number).category_id) for col in cols_of_cell]
                )
            else:
                col_number = 0

            raw_table_segments.append(
                SegmentationResult(
                    annotation_id=cell.annotation_id,
                    row_num=row_number,
                    col_num=col_number,
                    rs=rs,
                    cs=cs,
                )
            )

    if segment_mapping_context.context_error:
        return _default_segment_table(cells)
    return raw_table_segments


def create_intersection_cells(
    rows: List[ImageAnnotation],
    cols: List[ImageAnnotation],
    table_annotation_id: str,
    cell_class_id: int,
    sub_item_names: List[CellType],
) -> Tuple[List[DetectionResult], List[SegmentationResult]]:
    """
    Given rows and columns with row- and column number sub categories, create a list of `DetectionResult` and
    `SegmentationResult` as intersection of all their intersection rectangles.

    :param rows: list of rows
    :param cols: list of columns
    :param table_annotation_id: annotation_id of underlying table ImageAnnotation
    :param cell_class_id: The class_id to a synthetically generated DetectionResult
    :param sub_item_names: ObjectTypes for row-/column number
    :return: Pair of lists of `DetectionResult` and `SegmentationResult`.
    """
    boxes_rows = [row.image.get_embedding(table_annotation_id) for row in rows if row.image is not None]
    boxes_cols = [col.image.get_embedding(table_annotation_id) for col in cols if col.image is not None]

    boxes_cells = intersection_boxes(boxes_rows, boxes_cols)
    detect_result_cells = []
    segment_result_cells = []
    idx = 0
    for row in rows:
        for col in cols:
            detect_result_cells.append(
                DetectionResult(
                    box=boxes_cells[idx].to_list(mode="xyxy"),
                    class_id=cell_class_id,
                    absolute_coords=boxes_cells[idx].absolute_coords,
                    class_name=LayoutType.cell,
                )
            )
            segment_result_cells.append(
                SegmentationResult(
                    annotation_id="",
                    row_num=int(row.get_sub_category(sub_item_names[0]).category_id),
                    col_num=int(col.get_sub_category(sub_item_names[1]).category_id),
                    rs=1,
                    cs=1,
                )
            )
            idx += 1
    return detect_result_cells, segment_result_cells


def segment_pubtables(
    dp: Image,
    table: ImageAnnotation,
    item_names: List[LayoutType],
    spanning_cell_names: List[CellType],
    segment_rule: Literal["iou", "ioa"],
    threshold_rows: float,
    threshold_cols: float,
) -> List[SegmentationResult]:
    """
    Segment a table based on the results of `table-transformer-structure-recognition`. The processing assumes that cells
    have already been generated from the intersection of columns and rows and that column and row numbers have been
    inferred for rows and columns.

    Row and column positions as well as row and column lengths are determined for all types of spanning cells.
    All simple cells that are covered by a spanning cell as well in the table position (double allocation) are then
    removed.

    :param dp: Image
    :param table: table ImageAnnotation
    :param item_names: A list of item names (e.g. "row" and "column")
    :param spanning_cell_names: A list of spanning cell names (e.g. "projected_row_header" and "spanning")
    :param segment_rule: 'iou' or 'ioa'
    :param threshold_rows: the iou/ioa threshold of a cell with a row in order to conclude that the cell belongs
                               to the row.
    :param threshold_cols: the iou/ioa threshold of a cell with a column in order to conclude that the cell belongs
                               to the column.
    :return: A list of len(number of cells) of SegmentationResult for spanning cells
    """
    child_ann_ids = table.get_relationship(Relationships.child)
    cell_index_rows, row_index, _, _ = match_anns_by_intersection(
        dp,
        item_names[0],
        spanning_cell_names,
        segment_rule,
        threshold_rows,
        True,
        child_ann_ids,
        child_ann_ids,
    )

    cell_index_cols, col_index, _, _ = match_anns_by_intersection(
        dp,
        item_names[1],
        spanning_cell_names,
        segment_rule,
        threshold_cols,
        True,
        child_ann_ids,
        child_ann_ids,
    )

    spanning_cells = dp.get_annotation(annotation_ids=child_ann_ids, category_names=spanning_cell_names)

    rows = dp.get_annotation(annotation_ids=child_ann_ids, category_names=item_names[0])
    columns = dp.get_annotation(annotation_ids=child_ann_ids, category_names=item_names[1])

    raw_table_segments = []

    with MappingContextManager(dp_name=dp.file_name) as segment_mapping_context:
        for idx, cell in enumerate(spanning_cells):
            cell_positions_rows = cell_index_rows == idx
            rows_of_cell = [rows[k] for k in row_index[cell_positions_rows]]
            rs = np.count_nonzero(cell_index_rows == idx)
            if len(rows_of_cell):
                row_number = min([int(row.get_sub_category(CellType.row_number).category_id) for row in rows_of_cell])
            else:
                row_number = 0

            cell_positions_cols = cell_index_cols == idx
            cols_of_cell = [columns[k] for k in col_index[cell_positions_cols]]
            cs = np.count_nonzero(cell_index_cols == idx)
            if len(cols_of_cell):
                col_number = min(
                    [int(col.get_sub_category(CellType.column_number).category_id) for col in cols_of_cell]
                )
            else:
                col_number = 0

            raw_table_segments.append(
                SegmentationResult(
                    annotation_id=cell.annotation_id,
                    row_num=row_number,
                    col_num=col_number,
                    rs=rs,
                    cs=cs,
                )
            )

    if segment_mapping_context.context_error:
        return _default_segment_table(spanning_cells)
    return raw_table_segments


@pipeline_component_registry.register("TableSegmentationService")
class TableSegmentationService(PipelineComponent):
    """
    Table segmentation after successful cell detection. In addition, row and column detection must have been carried
    out.

    After cell recognition, these must be given a semantically correct position within the table. The row number,
    column number, row span and column span of the cell are determined. The determination takes place via an assignment
    via intersection.

    - Predicted rows are stretched horizontally to the edges of the table. Columns are stretched vertically. There is
      also the option of stretching rows and columns so that they completely pave the table (set tile_table_with_items
      =True).

    - Next, rows and columns are given a row or column number by sorting them vertically or horizontally
      according to the box center.

    - The averages are then determined in pairs separately for rows and columns (more precisely: Iou /
      intersection-over-union or ioa / intersection-over-area of rows and cells or columns and cells. A cell is
      assigned a row position if the iou / ioa is above a defined threshold.

    - The minimum row or column with which the cell was matched is used as the row and column of the cell. Row span /
      col span result from the number of matched rows and columns.

    It should be noted that this method means that cell positions can be assigned multiple times by different cells.
    If this should be excluded, `TableSegmentationRefinementService` can be used to merge cells.
    """

    def __init__(
        self,
        segment_rule: Literal["iou", "ioa"],
        threshold_rows: float,
        threshold_cols: float,
        tile_table_with_items: bool,
        remove_iou_threshold_rows: float,
        remove_iou_threshold_cols: float,
        stretch_rule: Literal["left", "equal"] = "left",
    ):
        """
        :param segment_rule: rule to assign cell to row, columns resp. must be either iou or ioa
        :param threshold_rows: iou/ioa threshold for rows
        :param threshold_cols: iou/ioa threshold for columns
        :param tile_table_with_items: Will shift the left edge of rows vertically to coincide with the right edge of
                                      the adjacent row. Will do a similar shifting with columns.
        :param remove_iou_threshold_rows: iou threshold for removing overlapping rows
        :param remove_iou_threshold_cols: iou threshold for removing overlapping columns
        :param stretch_rule: Check the description in `tile_tables_with_items_per_table`
        """
        assert segment_rule in ("iou", "ioa"), "segment_rule must be either iou or ioa"
        assert stretch_rule in ("left", "equal"), "stretch rule must be either 'left' or 'equal'"

        self.segment_rule = segment_rule
        self.threshold_rows = threshold_rows
        self.threshold_cols = threshold_cols
        self.tile_table = tile_table_with_items
        self.remove_iou_threshold_rows = remove_iou_threshold_rows
        self.remove_iou_threshold_cols = remove_iou_threshold_cols
        self.stretch_rule = stretch_rule
        self._table_name = LayoutType.table
        self._cell_names = [CellType.header, CellType.body, LayoutType.cell]
        self._item_names = [LayoutType.row, LayoutType.column]  # row names must be before column name
        self._sub_item_names = [CellType.row_number, CellType.column_number]
        super().__init__("table_segment")

    def serve(self, dp: Image) -> None:
        dp = stretch_items(
            dp,
            self._table_name,
            self._item_names[0],
            self._item_names[1],
            self.remove_iou_threshold_rows,
            self.remove_iou_threshold_cols,
        )
        table_anns = dp.get_annotation(category_names=self._table_name)
        for table in table_anns:
            item_ann_ids = table.get_relationship(Relationships.child)
            for item_sub_item_name in zip(self._item_names, self._sub_item_names):  # one pass for rows and one for cols
                item_name, sub_item_name = item_sub_item_name[0], item_sub_item_name[1]
                if self.tile_table:
                    dp = tile_tables_with_items_per_table(dp, table, item_name, self.stretch_rule)
                items_proposals = dp.get_annotation(category_names=item_name, annotation_ids=item_ann_ids)
                reference_items_proposals = dp.get_annotation(
                    category_names=self._cell_names, annotation_ids=item_ann_ids
                )
                dp = choose_items_by_iou(dp, items_proposals, 0.0001, False, reference_items_proposals)

                # new query of items so that we only get active annotations
                items = dp.get_annotation(category_names=item_name, annotation_ids=item_ann_ids)

                # we will assume that either all or no image attribute has been generated
                if items[0].image is not None:
                    items.sort(
                        key=lambda x: x.image.get_embedding(dp.image_id).cx  # type: ignore
                        if item_name == LayoutType.column  # pylint: disable=W0640
                        else x.image.get_embedding(dp.image_id).cy  # type: ignore
                    )
                else:
                    items.sort(
                        key=lambda x: x.bounding_box.cx  # type: ignore
                        if item_name == LayoutType.column  # pylint: disable=W0640
                        else x.bounding_box.cy  # type: ignore
                    )
                for item_number, item in enumerate(items, 1):
                    self.dp_manager.set_category_annotation(
                        sub_item_name, item_number, sub_item_name, item.annotation_id
                    )
            raw_table_segments = segment_table(
                dp,
                table,
                self._item_names,
                self._cell_names,
                self.segment_rule,
                self.threshold_rows,
                self.threshold_cols,
            )
            for segment_result in raw_table_segments:
                self.dp_manager.set_category_annotation(
                    CellType.row_number, segment_result.row_num, CellType.row_number, segment_result.annotation_id
                )
                self.dp_manager.set_category_annotation(
                    CellType.column_number, segment_result.col_num, CellType.column_number, segment_result.annotation_id
                )
                self.dp_manager.set_category_annotation(
                    CellType.row_span, segment_result.rs, CellType.row_span, segment_result.annotation_id
                )
                self.dp_manager.set_category_annotation(
                    CellType.column_span, segment_result.cs, CellType.column_span, segment_result.annotation_id
                )

    def clone(self) -> PipelineComponent:
        return self.__class__(
            self.segment_rule,
            self.threshold_rows,
            self.threshold_cols,
            self.tile_table,
            self.remove_iou_threshold_rows,
            self.remove_iou_threshold_cols,
        )

    def get_meta_annotation(self) -> JsonDict:
        return dict(
            [
                ("image_annotations", []),
                (
                    "sub_categories",
                    {
                        LayoutType.cell: {
                            CellType.row_number,
                            CellType.column_number,
                            CellType.row_span,
                            CellType.column_span,
                        },
                        LayoutType.row: {CellType.row_number},
                        LayoutType.column: {CellType.column_number},
                    },
                ),
                ("relationships", {}),
                ("summaries", []),
            ]
        )


class PubtablesSegmentationService(PipelineComponent):
    """
    Table segmentation for table recognition detectors trained on Pubtables1M dataset. It will require `ImageAnnotation`
    of type `LayoutType.row`, `LayoutType.column` and cells of at least one type `CellType.spanning`,
    `CellType.row_header`, `CellType.column_header`, `CellType.projected_row_header`. For table recognition using
    this service build a pipeline as follows:

    **Example:**

        layout = ImageLayoutService(layout_detector, to_image=True, crop_image=True)
        recognition = SubImageLayoutService(table_recognition_detector, LayoutType.table, {1: 6, 2:7, 3:8, 4:9}, True)
        segment = PubtablesSegmentationService('ioa', 0.4, 0.4, True, 0.8, 0.8, 7)
        ...

        pipe = DoctectionPipe([layout, recognition, segment])

    Under the hood this service performs the following tasks:

    - Stretching of rows and columns horizontally and vertically, so that the underlying table is fully tiled by rows
      and columns.
    - Enumerating rows and columns.
    - For intersecting rows and columns it will create an 'ImageAnnotation' of category 'LayoutType.cell'.
    - Using spanning cells from the detector to determine their 'row_number' and column_number' position.
    - Using cells and spanning cells, it will generate a tiling of the table with cells. When some cells have a position
      with some spanning cells, it will deactivate those simple cells and prioritize the spanning cells.
    - Determining the HTML representation of table.

    Different from the 'TableSegmentationService' this service does not require a refinement service: the advantage of
    this method is, that the segmentation can already be 'HTMLized'.
    """

    def __init__(
        self,
        segment_rule: Literal["iou", "ioa"],
        threshold_rows: float,
        threshold_cols: float,
        tile_table_with_items: bool,
        remove_iou_threshold_rows: float,
        remove_iou_threshold_cols: float,
        cell_class_id: int,
        cell_to_image: bool = True,
        crop_cell_image: bool = False,
        stretch_rule: Literal["left", "equal"] = "left",
    ) -> None:
        """

        :param segment_rule: rule to assign spanning cells to row, columns resp. must be either iou or ioa
        :param threshold_rows: iou/ioa threshold for rows
        :param threshold_cols: iou/ioa threshold for columns
        :param tile_table_with_items: Will shift the left edge of rows vertically to coincide with the right edge of
                                      the adjacent row. Will do a similar shifting with columns.
        :param remove_iou_threshold_rows: iou threshold for removing overlapping rows
        :param remove_iou_threshold_cols: iou threshold for removing overlapping columns
        :param cell_class_id: 'category_id' for cells to be generated from intersected rows and columns
        :param cell_to_image: If set to 'True' it will create an 'Image' for LayoutType.cell
        :param crop_cell_image: If set to 'True' it will crop a numpy array image for LayoutType.cell.
                                Requires 'cell_to_image=True'
        :param stretch_rule: Check the description in `tile_tables_with_items_per_table`
        """
        self.segment_rule = segment_rule
        self.threshold_rows = threshold_rows
        self.threshold_cols = threshold_cols
        self.tile_table = tile_table_with_items
        self._table_name = LayoutType.table
        self._cell_names = [
            CellType.spanning,
            CellType.row_header,
            CellType.column_header,
            CellType.projected_row_header,
            LayoutType.cell,
        ]
        self._spanning_cell_names = [
            CellType.spanning,
            CellType.row_header,
            CellType.column_header,
            CellType.projected_row_header,
        ]
        self.remove_iou_threshold_rows = remove_iou_threshold_rows
        self.remove_iou_threshold_cols = remove_iou_threshold_cols
        self.cell_class_id = cell_class_id
        self.cell_to_image = cell_to_image
        self.crop_cell_image = crop_cell_image
        self.stretch_rule = stretch_rule

        self._item_names = [LayoutType.row, LayoutType.column]  # row names must be before column name
        self._sub_item_names = [CellType.row_number, CellType.column_number]
        super().__init__("table_transformer_segment")

    def serve(self, dp: Image) -> None:
        dp = stretch_items(
            dp,
            self._table_name,
            self._item_names[0],
            self._item_names[1],
            self.remove_iou_threshold_rows,
            self.remove_iou_threshold_cols,
        )
        table_anns = dp.get_annotation(category_names=self._table_name)
        for table in table_anns:
            item_ann_ids = table.get_relationship(Relationships.child)
            for item_sub_item_name in zip(self._item_names, self._sub_item_names):  # one pass for rows and one for cols
                item_name, sub_item_name = item_sub_item_name[0], item_sub_item_name[1]
                if self.tile_table:
                    dp = tile_tables_with_items_per_table(dp, table, item_name, self.stretch_rule)
                items = dp.get_annotation(category_names=item_name, annotation_ids=item_ann_ids)

                # we will assume that either all or no image attribute has been generated
                if items[0].image is not None:
                    items.sort(
                        key=lambda x: x.image.get_embedding(dp.image_id).cx  # type: ignore
                        if item_name == LayoutType.column  # pylint: disable=W0640
                        else x.image.get_embedding(dp.image_id).cy  # type: ignore
                    )
                else:
                    items.sort(
                        key=lambda x: x.bounding_box.cx  # type: ignore
                        if item_name == LayoutType.column  # pylint: disable=W0640
                        else x.bounding_box.cy  # type: ignore
                    )

                for item_number, item in enumerate(items, 1):
                    self.dp_manager.set_category_annotation(
                        sub_item_name, item_number, sub_item_name, item.annotation_id
                    )
            rows = dp.get_annotation(category_names=self._item_names[0], annotation_ids=item_ann_ids)
            columns = dp.get_annotation(category_names=self._item_names[1], annotation_ids=item_ann_ids)
            detect_result_cells, segment_result_cells = create_intersection_cells(
                rows, columns, table.annotation_id, self.cell_class_id, self._sub_item_names
            )
            cell_rn_cn_to_ann_id = {}
            for detect_result, segment_result in zip(detect_result_cells, segment_result_cells):
                segment_result.annotation_id = self.dp_manager.set_image_annotation(  # type: ignore
                    detect_result,
                    to_annotation_id=table.annotation_id,
                    to_image=self.cell_to_image,
                    crop_image=self.crop_cell_image,
                )
                self.dp_manager.set_category_annotation(
                    CellType.row_number, segment_result.row_num, CellType.row_number, segment_result.annotation_id
                )
                self.dp_manager.set_category_annotation(
                    CellType.column_number, segment_result.col_num, CellType.column_number, segment_result.annotation_id
                )
                self.dp_manager.set_category_annotation(
                    CellType.row_span, segment_result.rs, CellType.row_span, segment_result.annotation_id
                )
                self.dp_manager.set_category_annotation(
                    CellType.column_span, segment_result.cs, CellType.column_span, segment_result.annotation_id
                )
                cell_rn_cn_to_ann_id[(segment_result.row_num, segment_result.col_num)] = segment_result.annotation_id
            spanning_cell_raw_segments = segment_pubtables(
                dp,
                table,
                self._item_names,
                self._spanning_cell_names,
                self.segment_rule,
                self.threshold_rows,
                self.threshold_cols,
            )
            for segment_result in spanning_cell_raw_segments:
                self.dp_manager.set_category_annotation(
                    CellType.row_number, segment_result.row_num, CellType.row_number, segment_result.annotation_id
                )
                self.dp_manager.set_category_annotation(
                    CellType.column_number, segment_result.col_num, CellType.column_number, segment_result.annotation_id
                )
                self.dp_manager.set_category_annotation(
                    CellType.row_span, segment_result.rs, CellType.row_span, segment_result.annotation_id
                )
                self.dp_manager.set_category_annotation(
                    CellType.column_span, segment_result.cs, CellType.column_span, segment_result.annotation_id
                )
                cells_to_deactivate = []
                for rs in range(segment_result.rs):
                    for cs in range(segment_result.cs):
                        cells_to_deactivate.append((segment_result.row_num + rs, segment_result.col_num + cs))
                for cell_position in cells_to_deactivate:
                    cell_ann_id = cell_rn_cn_to_ann_id[cell_position]
                    self.dp_manager.deactivate_annotation(cell_ann_id)

            cells = []
            if table.image:
                cells = table.image.get_annotation(category_names=self._cell_names)
            number_of_rows = max([int(cell.get_sub_category(CellType.row_number).category_id) for cell in cells])
            number_of_cols = max([int(cell.get_sub_category(CellType.column_number).category_id) for cell in cells])
            max_row_span = max([int(cell.get_sub_category(CellType.row_span).category_id) for cell in cells])
            max_col_span = max([int(cell.get_sub_category(CellType.column_span).category_id) for cell in cells])
            # TODO: the summaries should be sub categories of the underlying ann
            self.dp_manager.set_summary_annotation(
                TableType.number_of_rows, TableType.number_of_rows, number_of_rows, annotation_id=table.annotation_id
            )
            self.dp_manager.set_summary_annotation(
                TableType.number_of_columns,
                TableType.number_of_columns,
                number_of_cols,
                annotation_id=table.annotation_id,
            )
            self.dp_manager.set_summary_annotation(
                TableType.max_row_span, TableType.max_row_span, max_row_span, annotation_id=table.annotation_id
            )
            self.dp_manager.set_summary_annotation(
                TableType.max_col_span, TableType.max_col_span, max_col_span, annotation_id=table.annotation_id
            )
            html = generate_html_string(table)
            self.dp_manager.set_container_annotation(TableType.html, -1, TableType.html, table.annotation_id, html)

    def clone(self) -> PipelineComponent:
        return self.__class__(
            self.segment_rule,
            self.threshold_rows,
            self.threshold_cols,
            self.tile_table,
            self.remove_iou_threshold_rows,
            self.remove_iou_threshold_cols,
            self.cell_class_id,
            self.cell_to_image,
            self.crop_cell_image,
            self.stretch_rule,
        )

    def get_meta_annotation(self) -> JsonDict:
        return dict(
            [
                ("image_annotations", []),
                (
                    "sub_categories",
                    {
                        LayoutType.cell: {
                            CellType.row_number,
                            CellType.column_number,
                            CellType.row_span,
                            CellType.column_span,
                        },
                        CellType.spanning: {
                            CellType.row_number,
                            CellType.column_number,
                            CellType.row_span,
                            CellType.column_span,
                        },
                        CellType.row_header: {
                            CellType.row_number,
                            CellType.column_number,
                            CellType.row_span,
                            CellType.column_span,
                        },
                        CellType.column_header: {
                            CellType.row_number,
                            CellType.column_number,
                            CellType.row_span,
                            CellType.column_span,
                        },
                        CellType.projected_row_header: {
                            CellType.row_number,
                            CellType.column_number,
                            CellType.row_span,
                            CellType.column_span,
                        },
                        LayoutType.row: {CellType.row_number},
                        LayoutType.column: {CellType.column_number},
                    },
                ),
                ("relationships", {}),
                ("summaries", []),
            ]
        )



# -*- coding: utf-8 -*-
# File: logger.py

# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")
"""
This file is modified from
https://github.com/tensorpack/tensorpack/blob/master/tensorpack/utils/logger.py

The logger module itself has the common logging functions of Python's
`logging.Logger`.

**Example:**

    ..code-block::python

        from deepdoctection.utils.logger import logger

        logger.set_logger_dir("path/to/dir")
        logger.info("Something has happened")
        logger.warning("Attention!")
        logger.error("Error happened!")
"""

import errno
import functools
import json
import logging
import logging.config
import os
import shutil
import sys
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional, no_type_check

from termcolor import colored

from .detection_types import Pathlike

__all__ = ["logger", "set_logger_dir", "auto_set_dir", "get_logger_dir"]


class CustomFilter(logging.Filter):
    """A custom filter"""

    def filter(self, record: logging.LogRecord) -> bool:
        log_dict = {}
        args = record.args
        str_args = []
        if args:
            for arg in args:
                if isinstance(arg, dict):
                    log_dict.update(arg)
                else:
                    str_args.append(arg)
        record.args = tuple(str_args)
        if not hasattr(record, "log_dict"):
            setattr(record, "log_dict", log_dict)
        return True


class StreamFormatter(logging.Formatter):
    """A custom formatter to produce unified LogRecords"""

    @no_type_check
    def format(self, record: logging.LogRecord) -> str:
        date = colored("[%(asctime)s @%(filename)s:%(lineno)d]", "green")
        msg = colored("%(message)s", "white")

        if record.levelno == logging.WARNING:
            fmt = f"{date}  {colored('WRN', 'magenta', attrs=['blink'])}  {msg}"
        elif record.levelno == logging.ERROR or record.levelno == logging.CRITICAL:  # pylint: disable=R1714
            fmt = f"{date}  {colored('ERR', 'red', attrs=['blink', 'underline'])}  {msg}"
        elif record.levelno == logging.DEBUG:
            fmt = f"{date}  {colored('DBG', 'green', attrs=['blink'])}  {msg}"
        elif record.levelno == logging.INFO:
            fmt = f"{date}  {colored('INF', 'green')}  {msg}"
        else:
            fmt = f"{date} {msg}"
        self._style._fmt = fmt  # pylint: disable=W0212
        self._fmt = fmt
        return super().format(record)


class FileFormatter(logging.Formatter):
    """A custom formatter to produce a loggings in json format"""

    @no_type_check
    def format(self, record: logging.LogRecord) -> str:
        message = super().format(record)
        log_dict = {
            "level_no": record.levelno,
            "level_name": record.levelname,
            "module_name": record.filename,
            "line_number": record.lineno,
            "time": datetime.now().strftime("%m%d-%H%M%S"),
            "message": message,
        }
        log_dict.update(record.log_dict)
        return json.dumps(log_dict)


_LOG_DIR = None
_CONFIG_DICT: Dict[str, Any] = {
    "version": 1,
    "filters": {"customfilter": {"()": lambda: CustomFilter()}},  # pylint: disable=W0108
    "formatters": {
        "streamformatter": {"()": lambda: StreamFormatter(datefmt="%m%d %H:%M.%S")},
    },
    "handlers": {
        "streamhandler": {"filters": ["customfilter"], "formatter": "streamformatter", "class": "logging.StreamHandler"}
    },
    "root": {"handlers": ["streamhandler"], "level": "INFO", "propagate": False},
}


def _get_logger() -> logging.Logger:
    logging.config.dictConfig(_CONFIG_DICT)
    _logger = logging.getLogger(__name__)
    return _logger


logger = _get_logger()

_LOGGING_METHOD = ["info", "warning", "error", "critical", "debug"]
for func in _LOGGING_METHOD:
    locals()[func] = getattr(logger, func)
    __all__.append(func)


_FILE_HANDLER = None


def _get_time_str() -> str:
    return datetime.now().strftime("%m%d-%H%M%S")


def _set_file(path: Pathlike) -> None:
    if isinstance(path, Path):
        path = path.as_posix()
    global _FILE_HANDLER  # pylint: disable=W0603
    if os.path.isfile(path):
        backup_name = path + "." + _get_time_str()
        shutil.move(path, backup_name)
        logger.info("Existing log file %s backuped to %s", path, backup_name)
    hdl = logging.FileHandler(filename=path, encoding="utf-8", mode="w")
    hdl.setFormatter(FileFormatter(datefmt="%m%d %H:%M:%S"))
    hdl.addFilter(CustomFilter())

    _FILE_HANDLER = hdl
    logger.addHandler(hdl)
    logger.info("Argv: %s ", sys.argv)


def set_logger_dir(dir_name: Pathlike, action: Optional[str] = None) -> None:
    """
    Set the directory for global logging.

    :param dir_name: log directory
    :param action: an action of ["k","d","q"] to be performed
                   when the directory exists. Will ask user by default.

                   "d": delete the directory. Note that the deletion may fail when
                   the directory is used by tensorboard.
                   "k": keep the directory. This is useful when you resume from a
                   previous training and want the directory to look as if the
                   training was not interrupted.
                   Note that this option does not load old models or any other
                   old states for you. It simply does nothing.
    """
    if isinstance(dir_name, Path):
        dir_name = dir_name.as_posix()
    dir_name = os.path.normpath(dir_name)
    global _LOG_DIR, _FILE_HANDLER  # pylint: disable=W0603
    if _FILE_HANDLER:
        # unload and close the old file handler, so that we may safely delete the logger directory
        logger.removeHandler(_FILE_HANDLER)
        del _FILE_HANDLER

    def dir_nonempty(directory: str) -> int:
        return os.path.isdir(directory) and len([x for x in os.listdir(directory) if x[0] != "."])

    if dir_nonempty(dir_name):
        if not action:
            logger.warning("Log directory %s exists! Use 'd' to delete it. ", dir_name)
            logger.warning(
                "If you're resuming from a previous run, you can choose to keep it." "Press any other key to exit. "
            )
        while not action:
            action = input("Select Action: k (keep) / d (delete) / q (quit):").lower().strip()
        act = action
        if act == "b":
            backup_name = dir_name + _get_time_str()
            shutil.move(dir_name, backup_name)
            logger.info("Directory %s backuped to %s", dir_name, backup_name)
        elif act == "d":
            shutil.rmtree(dir_name, ignore_errors=True)
            if dir_nonempty(dir_name):
                shutil.rmtree(dir_name, ignore_errors=False)
        elif act == "n":
            dir_name = dir_name + _get_time_str()
            logger.info("Use a new log directory %s ", dir_name)
        elif act == "k":
            pass
        else:
            raise OSError(f"Directory {dir_name} exits!")
    _LOG_DIR = os.path.join(dir_name, "log.jsonl")
    try:
        os.makedirs(dir_name)
    except OSError as err:
        if err.errno != errno.EEXIST:
            raise err

    _set_file(_LOG_DIR)


def auto_set_dir(action: Optional[str] = None, name: Optional[str] = None) -> None:
    """
    Will set the log directory to './train_log/{script_name}:{name}'.
    'script_name' is the name of the main python file currently running.

    :param action: an action of ["k","d","q"] to be performed (see also func: set_logger_dir)
    :param name: Optional suffix of file name.
    """

    mod = sys.modules["__main__"]
    basename = str(os.path.basename(mod.__file__))  # type: ignore
    auto_dir_name = os.path.join("train_log", basename[: basename.rfind(".")])
    if name:
        auto_dir_name += "_%s" % name if os.name == "nt" else ":%s" % name  # pylint: disable=C0209
    set_logger_dir(auto_dir_name, action=action)


def get_logger_dir() -> Optional[str]:
    """
    The logger directory, or None if not set.
    The directory is used for general logging, tensorboard events, checkpoints, etc.
    """
    return _LOG_DIR


@functools.lru_cache(maxsize=None)
def log_once(message: str, function: str = "info") -> None:
    """
    Log certain message only once. Call this function more than one times with
    the same message will result in no-op.

    :param message:  message to log
    :param function: the name of the logger method. e.g. "info", "warn", "error".
    :return:
    """
    getattr(logger, function)(message)



# -*- coding: utf-8 -*-
# File: test_cell.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module pipe.cell
"""

from typing import List
from unittest.mock import MagicMock

from pytest import mark

from deepdoctection.datapoint import BoundingBox, Image
from deepdoctection.datasets import DatasetCategories
from deepdoctection.extern.base import DetectionResult
from deepdoctection.pipe.base import ObjectDetector
from deepdoctection.pipe.cell import DetectResultGenerator, SubImageLayoutService
from deepdoctection.utils.settings import LayoutType


@mark.basic
def test_detect_result_generator(
    dataset_categories: DatasetCategories, dp_image: Image, layout_detect_results: List[DetectionResult]
) -> None:
    """
    Testing DetectResultGenerator creates DetectionResult correctly
    """

    # Arrange
    categories = dataset_categories.get_categories()
    detect_result_generator = DetectResultGenerator(categories, [["1"], ["2"], ["3"], ["4"], ["5"]])

    # Act
    detect_result_generator.width = 600
    detect_result_generator.height = 400
    raw_anns = detect_result_generator.create_detection_result(layout_detect_results)

    # Assert
    raw_ann_cats = {raw_ann.class_id for raw_ann in raw_anns}
    assert raw_ann_cats == {1, 2, 3, 4, 5}

    assert raw_anns[5].box == [0.0, 0.0, dp_image.width, dp_image.height]


class TestSubImageLayoutService:
    """
    Test SubImageLayoutService
    """

    def setup_method(self) -> None:
        """
        setup necessary components
        """

        self._cell_detector = MagicMock(spec=ObjectDetector)
        self._cell_detector.name = "mock_cell_detector"

        self.sub_image_layout_service = SubImageLayoutService(self._cell_detector, LayoutType.table)

    @mark.basic
    def test_pass_datapoint(
        self,
        dp_image_with_layout_anns: Image,
        cell_detect_results: List[List[DetectionResult]],
        global_cell_boxes: List[List[BoundingBox]],
    ) -> None:
        """
        Testing pass_datapoint
        """

        # Arrange
        self._cell_detector.predict = MagicMock(side_effect=cell_detect_results)

        # Act
        dp = self.sub_image_layout_service.pass_datapoint(dp_image_with_layout_anns)
        anns = dp.get_annotation(category_names=LayoutType.table)

        # Assert
        assert len(anns) == 2

        first_table_ann = anns[0]
        second_table_ann = anns[1]

        exp_global_boxes_first_table = global_cell_boxes[0]
        exp_global_boxes_scd_table = global_cell_boxes[1]

        first_table_cell_anns = first_table_ann.image.get_annotation()  # type: ignore
        second_table_cell_anns = second_table_ann.image.get_annotation()  # type: ignore
        assert len(first_table_cell_anns) == 2
        assert len(second_table_cell_anns) == 1

        first_table_first_cell = first_table_cell_anns[0]
        first_table_second_cell = first_table_cell_anns[1]
        second_table_first_cell = second_table_cell_anns[0]

        global_box_ftfc = first_table_first_cell.image.get_embedding(dp.image_id)
        assert global_box_ftfc == exp_global_boxes_first_table[0]
        local_box_ftfc = first_table_first_cell.image.get_embedding(first_table_ann.annotation_id)
        assert local_box_ftfc == first_table_first_cell.bounding_box

        global_box_ftsc = first_table_second_cell.image.get_embedding(dp.image_id)
        assert global_box_ftsc == exp_global_boxes_first_table[1]
        local_box_ftsc = first_table_second_cell.image.get_embedding(first_table_ann.annotation_id)
        assert local_box_ftsc == first_table_second_cell.bounding_box

        global_box_stfc = second_table_first_cell.image.get_embedding(dp.image_id)
        assert global_box_stfc == exp_global_boxes_scd_table[0]
        local_box_stfc = second_table_first_cell.image.get_embedding(second_table_ann.annotation_id)
        assert local_box_stfc == second_table_first_cell.bounding_box



# -*- coding: utf-8 -*-
# File: conftest.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Fixtures for dataflow package testing
"""

from dataclasses import dataclass
from typing import List

from pytest import fixture


@fixture(name="datapoint_list")
def fixture_datapoint_list() -> List[str]:
    """
    List fixture
    """
    return ["a", "b", "c", "d"]


@dataclass
class DatasetThreeDim:
    """
    Dataset of shape (2,2,3)
    """

    dataset = [[[1.0, 0.0, 2.0], [0.5, 1.0, 1.0]], [[2.0, 0.0, 4.0], [0.0, 1.0, 1.0]]]
    mean_axis_zero = [[1.5, 0.0, 3.0], [0.25, 1.0, 1.0]]
    mean_all_axes = 1.125
    std_axis_zero = [[0.70710678, 0.0, 1.41421356], [0.35355339, 0.0, 0.0]]
    std_all_axes = 0.29462782549439476
    dataset_flatten = [[[1.0, 0.0, 2.0]], [[0.5, 1.0, 1.0]], [[2.0, 0.0, 4.0]], [[0.0, 1.0, 1.0]]]


@fixture(name="dataset_three_dim")
def fixture_dataset_three_dim() -> List[List[List[float]]]:
    """
    Dataset fixture
    """
    return DatasetThreeDim().dataset


@fixture(name="mean_axis_zero")
def fixture_mean_axis_zero() -> List[List[float]]:
    """
    Mean fixture
    """
    return DatasetThreeDim().mean_axis_zero


@fixture(name="mean_all_axes")
def fixture_mean_all_axes() -> float:
    """
    Mean fixture
    """
    return DatasetThreeDim().mean_all_axes


@fixture(name="std_axis_zero")
def fixture_std_axis_zero() -> List[List[float]]:
    """
    Std fixture
    """
    return DatasetThreeDim().std_axis_zero


@fixture(name="std_all_axes")
def fixture_std_all_axes() -> float:
    """
    Std fixture
    """
    return DatasetThreeDim().std_all_axes


@fixture(name="dataset_flatten")
def fixture_dataset_flatten() -> List[List[List[float]]]:
    """
    Dataset flatten
    """
    return DatasetThreeDim().dataset_flatten



# -*- coding: utf-8 -*-
# File: settings.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for funcs and constants that maintain general settings
"""

import os
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import catalogue  # type: ignore


class ObjectTypes(str, Enum):
    """Base Class for describing objects as attributes of Enums"""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}.{self.name}>"

    @classmethod
    def from_value(cls, value: str) -> "ObjectTypes":
        """Getting the enum member from a given string value

        :param value: string value to get the enum member
        :return: Enum member
        """
        for member in cls.__members__.values():
            if member.value == value:
                return member
        raise ValueError(f"value {value} does not have corresponding member")


TypeOrStr = Union[ObjectTypes, str]

object_types_registry = catalogue.create("deepdoctection", "settings", entry_points=True)


# pylint: disable=invalid-name
@object_types_registry.register("DefaultType")
class DefaultType(ObjectTypes):
    """Type for default member"""

    default_type = "default_type"


@object_types_registry.register("PageType")
class PageType(ObjectTypes):
    """Type for document page properties"""

    document_type = "document_type"
    language = "language"


@object_types_registry.register("SummaryType")
class SummaryType(ObjectTypes):
    """Summary type member"""

    summary = "summary"


@object_types_registry.register("DocumentType")
class DocumentType(ObjectTypes):
    """Document types"""

    letter = "letter"
    form = "form"
    email = "email"
    handwritten = "handwritten"
    advertisement = "advertisement"
    scientific_report = "scientific_report"
    scientific_publication = "scientific_publication"
    specification = "specification"
    file_folder = "file_folder"
    news_article = "news_article"
    budget = "budget"
    invoice = "invoice"
    presentation = "presentation"
    questionnaire = "questionnaire"
    resume = "resume"
    memo = "memo"
    financial_report = "financial_report"
    laws_and_regulations = "laws_and_regulations"
    government_tenders = "government_tenders"
    manuals = "manuals"
    patents = "patents"


@object_types_registry.register("LayoutType")
class LayoutType(ObjectTypes):
    """Layout types"""

    table = "table"
    table_rotated = "table_rotated"
    figure = "figure"
    list = "list"
    text = "text"
    title = "title"  # type: ignore
    logo = "logo"
    signature = "signature"
    caption = "caption"
    footnote = "footnote"
    formula = "formula"
    page_footer = "page_footer"
    page_header = "page_header"
    section_header = "section_header"
    page = "page"
    cell = "cell"
    row = "row"
    column = "column"
    word = "word"
    line = "line"


@object_types_registry.register("TableType")
class TableType(ObjectTypes):
    """Types for table properties"""

    item = "item"
    number_of_rows = "number_of_rows"
    number_of_columns = "number_of_columns"
    max_row_span = "max_row_span"
    max_col_span = "max_col_span"
    html = "html"


@object_types_registry.register("CellType")
class CellType(ObjectTypes):
    """Types for cell properties"""

    header = "header"
    body = "body"
    row_number = "row_number"
    row_span = "row_span"
    row_header = "row_header"
    projected_row_header = "projected_row_header"
    column_number = "column_number"
    column_span = "column_span"
    column_header = "column_header"
    spanning = "spanning"


@object_types_registry.register("WordType")
class WordType(ObjectTypes):
    """Types for word properties"""

    characters = "characters"
    block = "block"
    token_class = "token_class"
    tag = "tag"
    token_tag = "token_tag"
    text_line = "text_line"


@object_types_registry.register("TokenClasses")
class TokenClasses(ObjectTypes):
    """Types for token classes"""

    header = "header"
    question = "question"
    answer = "answer"
    other = "other"


@object_types_registry.register("BioTag")
class BioTag(ObjectTypes):
    """Types for tags"""

    begin = "B"
    inside = "I"
    outside = "O"
    single = "S"
    end = "E"


@object_types_registry.register("TokenClassWithTag")
class TokenClassWithTag(ObjectTypes):
    """Types for token classes with tags, e.g. B-answer"""

    b_answer = "B-answer"
    b_header = "B-header"
    b_question = "B-question"
    e_answer = "E-answer"
    e_header = "E-header"
    e_question = "E-question"
    i_answer = "I-answer"
    i_header = "I-header"
    i_question = "I-question"
    s_answer = "S-answer"
    s_header = "S-header"
    s_question = "S-question"


@object_types_registry.register("Relationships")
class Relationships(ObjectTypes):
    """Types for describing relationships between types"""

    child = "child"
    reading_order = "reading_order"
    semantic_entity_link = "semantic_entity_link"


@object_types_registry.register("Languages")
class Languages(ObjectTypes):
    """Language types"""

    english = "eng"
    russian = "rus"
    german = "deu"
    french = "fre"
    italian = "ita"
    japanese = "jpn"
    spanish = "spa"
    cebuano = "ceb"
    turkish = "tur"
    portuguese = "por"
    ukrainian = "ukr"
    esperanto = "epo"
    polish = "pol"
    swedish = "swe"
    dutch = "dut"
    hebrew = "heb"
    chinese = "chi"
    hungarian = "hun"
    arabic = "ara"
    catalan = "cat"
    finnish = "fin"
    czech = "cze"
    persian = "per"
    serbian = "srp"
    greek = "gre"
    vietnamese = "vie"
    bulgarian = "bul"
    korean = "kor"
    norwegian = "nor"
    macedonian = "mac"
    romanian = "rum"
    indonesian = "ind"
    thai = "tha"
    armenian = "arm"
    danish = "dan"
    tamil = "tam"
    hindi = "hin"
    croatian = "hrv"
    belarusian = "bel"
    georgian = "geo"
    telugu = "tel"
    kazakh = "kaz"
    waray = "war"
    lithuanian = "lit"
    scottish = "glg"
    slovak = "slo"
    benin = "ben"
    basque = "baq"
    slovenian = "slv"
    malayalam = "mal"
    marathi = "mar"
    estonian = "est"
    azerbaijani = "aze"
    albanian = "alb"
    latin = "lat"
    bosnian = "bos"
    norwegian_nynorsk = "nno"
    urdu = "urd"
    not_defined = "nn"


@object_types_registry.register("DatasetType")
class DatasetType(ObjectTypes):
    """Dataset types"""

    object_detection = "object_detection"
    sequence_classification = "sequence_classification"
    token_classification = "token_classification"
    publaynet = "publaynet"


_TOKEN_AND_TAG_TO_TOKEN_CLASS_WITH_TAG = {
    (TokenClasses.header, BioTag.begin): TokenClassWithTag.b_header,
    (TokenClasses.header, BioTag.inside): TokenClassWithTag.i_header,
    (TokenClasses.header, BioTag.end): TokenClassWithTag.e_header,
    (TokenClasses.header, BioTag.single): TokenClassWithTag.s_header,
    (TokenClasses.answer, BioTag.begin): TokenClassWithTag.b_answer,
    (TokenClasses.answer, BioTag.inside): TokenClassWithTag.i_answer,
    (TokenClasses.answer, BioTag.end): TokenClassWithTag.e_answer,
    (TokenClasses.answer, BioTag.single): TokenClassWithTag.s_answer,
    (TokenClasses.question, BioTag.begin): TokenClassWithTag.b_question,
    (TokenClasses.question, BioTag.inside): TokenClassWithTag.i_question,
    (TokenClasses.question, BioTag.end): TokenClassWithTag.e_question,
    (TokenClasses.question, BioTag.single): TokenClassWithTag.s_question,
    (TokenClasses.other, BioTag.outside): BioTag.outside,
    (TokenClasses.header, BioTag.outside): BioTag.outside,
    (TokenClasses.answer, BioTag.outside): BioTag.outside,
    (TokenClasses.question, BioTag.outside): BioTag.outside,
}


def token_class_tag_to_token_class_with_tag(token: ObjectTypes, tag: ObjectTypes) -> ObjectTypes:
    """
    Mapping TokenClassWithTag enum member from token class and tag, e.g. `TokenClasses.header` and `BioTag.inside`
    maps to TokenClassWithTag.i_header.

    :param token: TokenClasses member
    :param tag: BioTag member
    :return: TokenClassWithTag member
    """
    if isinstance(token, TokenClasses) and isinstance(tag, BioTag):
        return _TOKEN_AND_TAG_TO_TOKEN_CLASS_WITH_TAG[(token, tag)]
    raise TypeError("Token must be of type TokenClasses and tag must be of type BioTag")


def token_class_with_tag_to_token_class_and_tag(
    token_class_with_tag: ObjectTypes,
) -> Optional[Tuple[ObjectTypes, ObjectTypes]]:
    """
    This is the reverse mapping from TokenClassWithTag members to TokenClasses and BioTag

    :param token_class_with_tag: TokenClassWithTag member
    :return: Tuple of TokenClasses member and BioTag member
    """
    return {val: key for key, val in _TOKEN_AND_TAG_TO_TOKEN_CLASS_WITH_TAG.items()}.get(token_class_with_tag)


_ALL_TYPES_DICT = {}
_ALL_TYPES = set(object_types_registry.get_all().values())
for ob in _ALL_TYPES:
    _ALL_TYPES_DICT.update({e.value: e for e in ob})


def update_all_types_dict() -> None:
    """Updates subsequently registered object types. Useful for defining additional ObjectTypes in tests"""
    maybe_new_types = set(object_types_registry.get_all().values())
    difference = maybe_new_types - _ALL_TYPES
    for obj in difference:
        _ALL_TYPES_DICT.update({e.value: e for e in obj})


_OLD_TO_NEW_OBJ_TYPE: Dict[str, str] = {
    "DOC_CLASS": "document_type",
    "CHARS": "characters",
    "BIO_TAG": "tag",
    "B-ANSWER": "B-answer",
    "B-HEADER": "B-header",
    "B-QUESTION": "B-question",
    "E-ANSWER": "E-answer",
    "E-HEADER": "E-header",
    "E-QUESTION": "E-question",
    "I-ANSWER": "I-answer",
    "I-HEADER": "I-header",
    "I-QUESTION": "I-question",
    "S-ANSWER": "S-answer",
    "S-HEADER": "S-header",
    "S-QUESTION": "S-question",
}


def _get_new_obj_type_str(obj_type: str) -> str:
    return _OLD_TO_NEW_OBJ_TYPE.get(obj_type, obj_type)


_BLACK_LIST: List[str] = ["B", "I", "O", "E", "S"]


def _get_black_list() -> List[str]:
    return _BLACK_LIST


def update_black_list(item: str) -> None:
    """Updates the black list, i.e. set of elements that must not be lowered"""
    _BLACK_LIST.append(item)


def get_type(obj_type: Union[str, ObjectTypes]) -> ObjectTypes:
    """Get an object type property from a given string. Does nothing if an ObjectType is passed

    :param obj_type: String or ObjectTypes
    :return: ObjectType
    """
    if isinstance(obj_type, ObjectTypes):
        return obj_type
    obj_type = _get_new_obj_type_str(obj_type)
    if obj_type.startswith(("B-", "E-", "I-", "S-")):
        obj_type = obj_type[:2] + obj_type[2:].lower()
    elif obj_type not in _get_black_list():
        obj_type = obj_type.lower()
    return_value = _ALL_TYPES_DICT.get(obj_type)
    if return_value is None:
        raise KeyError(f"String {obj_type} does not correspond to a registered ObjectType")
    return return_value


# Some path settings

# package path
file_path = Path(os.path.split(__file__)[0])
PATH = file_path.parent.parent

# model cache directory
dd_cache_home = Path(os.getenv("XDG_CACHE_HOME", Path.home() / ".cache")) / "deepdoctection"
MODEL_DIR = dd_cache_home / "weights"

# configs cache directory
CONFIGS = dd_cache_home / "configs"

# dataset cache directory
DATASET_DIR = dd_cache_home / "datasets"

FILE_PATH = os.path.split(__file__)[0]
TPATH = os.path.dirname(os.path.dirname(FILE_PATH))



# -*- coding: utf-8 -*-
# File: iiitarstruct.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for mapping annotations in iiitar13k style structure
"""

import os
from typing import Dict, Optional

from ..datapoint.annotation import ImageAnnotation
from ..datapoint.box import BoundingBox
from ..datapoint.image import Image
from ..utils.detection_types import JsonDict
from ..utils.fs import load_image_from_file
from ..utils.settings import get_type
from .maputils import MappingContextManager, curry, maybe_get_fake_score


@curry
def pascal_voc_dict_to_image(
    dp: JsonDict,
    categories_name_as_key: Dict[str, str],
    load_image: bool,
    filter_empty_image: bool,
    fake_score: bool,
    category_name_mapping: Optional[Dict[str, str]] = None,
) -> Optional[Image]:
    """
    Map a dataset in a structure equivalent to iiitar13k annotation style to image format

    :param dp: a datapoint in serialized iiitar13k format. Note that another conversion from xml to
               a dict structure is required.
    :param categories_name_as_key: A dict of categories, e.g. DatasetCategories.get_categories(name_as_key=True)
    :param load_image: If 'True' it will load image to attr: Image.image
    :param filter_empty_image: Will return None, if datapoint has no annotations
    :param fake_score: If dp does not contain a score, a fake score with uniform random variables in (0,1)
                       will be added.
    :param category_name_mapping: Map incoming category names, e.g. {"source_name":"target_name"}
    :return: Image
    """

    anns = dp.get("objects", [])
    if not anns and filter_empty_image:
        return None

    with MappingContextManager(dp.get("filename")) as mapping_context:
        image = Image(
            file_name=os.path.split(dp["filename"])[1].replace(".xml", ".jpg"),
            location=dp["filename"].replace(".xml", ".jpg").replace("xml", "images"),
        )

        if load_image:
            image.image = load_image_from_file(image.location)
        image.set_width_height(float(dp.get("width", 0)), float(dp.get("height", 0)))

        for ann in anns:
            x_1 = min(max(ann["xmin"], 0), image.width if image.width else float(dp.get("width", 0)))
            x_2 = min(max(ann["xmax"], 0), image.width if image.width else float(dp.get("width", 0)))
            y_1 = min(max(ann["ymin"], 0), image.height if image.height else float(dp.get("height", 0)))
            y_2 = min(max(ann["ymax"], 0), image.height if image.height else float(dp.get("height", 0)))

            bbox = BoundingBox(absolute_coords=True, ulx=x_1, uly=y_1, lrx=x_2, lry=y_2)

            if category_name_mapping is not None:
                label = category_name_mapping.get(ann["name"])
                if not label:
                    label = ann["name"]
            else:
                label = ann["label"]
            assert isinstance(label, str)

            annotation = ImageAnnotation(
                category_name=get_type(label),
                bounding_box=bbox,
                category_id=categories_name_as_key[label],
                score=maybe_get_fake_score(fake_score),
            )
            image.dump(annotation)

    if mapping_context.context_error:
        return None
    return image



# -*- coding: utf-8 -*-
# File: __init__.py

"""
Dataflow is a package  for loading and processing data in both training and prediction environments. Dataflow
is essentially pure Python and, with a simple API, it contains a variety of methods for parallelling complex
transformations. Due to the fact that DataFlow is not available through pypi we have integrated the most important
DataFlow classes into deepdoctection in order to avoid installing the package separately from source.

Further information (including several tutorials about performance) can be found in the excellent documentation:

<https://tensorpack.readthedocs.io/en/latest/tutorial/dataflow.html>
"""


from .base import *
from .common import *
from .custom import *
from .custom_serialize import *
from .parallel_map import *
from .serialize import *
from .stats import *



# -*- coding: utf-8 -*-
# File: eval.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module for `Evaluator`
"""

__all__ = ["Evaluator"]

from copy import deepcopy
from typing import Any, Dict, List, Literal, Mapping, Optional, Type, Union, overload

import numpy as np

from ..dataflow import CacheData, DataFlow, DataFromList, MapData
from ..datapoint.image import Image
from ..datasets.base import DatasetBase
from ..mapper.cats import filter_cat, remove_cats
from ..mapper.misc import maybe_load_image, maybe_remove_image, maybe_remove_image_from_category
from ..pipe.base import LanguageModelPipelineComponent, PredictorPipelineComponent
from ..pipe.common import PageParsingService
from ..pipe.concurrency import MultiThreadPipelineComponent
from ..pipe.doctectionpipe import DoctectionPipe
from ..utils.detection_types import ImageType
from ..utils.file_utils import detectron2_available, wandb_available
from ..utils.logger import logger
from ..utils.settings import DatasetType, LayoutType, TypeOrStr
from ..utils.viz import interactive_imshow
from .base import MetricBase

if wandb_available():
    import wandb  # pylint:disable=W0611
    from wandb import Artifact, Table

if wandb_available() and detectron2_available():
    from ..mapper.d2struct import to_wandb_image


class Evaluator:
    """
    The API for evaluating pipeline components or pipelines on a given dataset. For a given model, a given dataset and
    a given metric, this class will stream the dataset, call the predictor(s) and will evaluate the predictions against
    the ground truth with respect to the given metric.

    After initializing the evaluator the process itself will start after calling the `run`.

    The following takes place under the hood:

    Setup of the dataflow according to the build- and split inputs. The `datasets.DataFlowBaseBuilder.build` will
    be invoked twice as one dataflow must be kept with its ground truth while the other must go through an annotation
    erasing process and after that passing the predictor. Predicted and gt datapoints will be converted into the
    required metric input format and dumped into lists. Both lists will be passed to `MetricBase.get_distance`.

    **Example:**

        You can evaluate the predictor on a subset of categories by filtering the ground truth dataset. When using
        the coco metric all predicted objects that are not in the set of filtered objects will be not taken into
        account.

            publaynet = get_dataset("publaynet")
            publaynet.dataflow.categories.filter_categories(categories=["TEXT","TITLE"])
            coco_metric = metric_registry.get("coco")
            profile = ModelCatalog.get_profile("layout/d2_model_0829999_layout_inf_only.pt")
            path_weights = ModelCatalog.get_full_path_weights("layout/d2_model_0829999_layout_inf_only.pt")
            path_config_yaml= ModelCatalog.get_full_path_configs("layout/d2_model_0829999_layout_inf_only.pt")

            layout_detector = D2FrcnnDetector(path_config_yaml, path_weights, profile.categories)
            layout_service = ImageLayoutService(layout_detector)
            evaluator = Evaluator(publaynet, layout_service, coco_metric)

            output = evaluator.run(max_datapoints=10)

    For another example check the script in :ref:`Evaluation of table recognition`
    """

    def __init__(
        self,
        dataset: DatasetBase,
        component_or_pipeline: Union[PredictorPipelineComponent, LanguageModelPipelineComponent, DoctectionPipe],
        metric: Union[Type[MetricBase], MetricBase],
        num_threads: int = 2,
        run: Optional["wandb.sdk.wandb_run.Run"] = None,
    ) -> None:
        """
        Evaluating a pipeline component on a dataset with a given metric.

        :param dataset: dataset
        :param component_or_pipeline: A pipeline component with predictor and annotation factory.
        :param metric: metric
        """

        self.dataset = dataset
        self.pipe_component: Optional[MultiThreadPipelineComponent] = None
        self.pipe: Optional[DoctectionPipe] = None

        # when passing a component, we will process prediction on num_threads
        if isinstance(component_or_pipeline, (PredictorPipelineComponent, LanguageModelPipelineComponent)):
            logger.info(
                "Building multi threading pipeline component to increase prediction throughput. Using %i threads",
                num_threads,
            )
            pipeline_components: List[Union[PredictorPipelineComponent, LanguageModelPipelineComponent]] = []

            for _ in range(num_threads - 1):
                copy_pipe_component = component_or_pipeline.clone()
                pipeline_components.append(copy_pipe_component)

            pipeline_components.append(component_or_pipeline)

            self.pipe_component = MultiThreadPipelineComponent(
                pipeline_components=pipeline_components,
                pre_proc_func=maybe_load_image,
                post_proc_func=maybe_remove_image,
            )
        else:
            self.pipe = component_or_pipeline

        self.metric = metric
        if not isinstance(metric, MetricBase):
            self.metric = self.metric()  # type: ignore

        self._sanity_checks()

        self.wandb_table_agent: Optional[WandbTableAgent]
        if run is not None:
            self.wandb_table_agent = WandbTableAgent(
                run, self.dataset.dataset_info.name, 50, self.dataset.dataflow.categories.get_categories(filtered=True)
            )
        else:
            self.wandb_table_agent = None

    @overload
    def run(
        self, output_as_dict: Literal[False] = False, **dataflow_build_kwargs: Union[str, int]
    ) -> List[Dict[str, float]]:
        ...

    @overload
    def run(self, output_as_dict: Literal[True], **dataflow_build_kwargs: Union[str, int]) -> Dict[str, float]:
        ...

    def run(
        self, output_as_dict: bool = False, **dataflow_build_kwargs: Union[str, int]
    ) -> Union[List[Dict[str, float]], Dict[str, float]]:
        """
        Start evaluation process and return the results.

        :param output_as_dict: Return result in a list or dict.
        :param dataflow_build_kwargs: Pass the necessary arguments in order to build the dataflow, e.g. "split",
                                      "build_mode", "max_datapoints" etc.

        :return: dict with metric results.
        """

        df_gt = self.dataset.dataflow.build(**dataflow_build_kwargs)
        df_pr = self.dataset.dataflow.build(**dataflow_build_kwargs)

        df_pr = MapData(df_pr, deepcopy)
        df_pr = self._clean_up_predict_dataflow_annotations(df_pr)
        df_pr = self._run_pipe_or_component(df_pr)

        logger.info("Starting evaluation...")
        result = self.metric.get_distance(df_gt, df_pr, self.dataset.dataflow.categories)
        self.metric.print_result()

        if self.wandb_table_agent:
            self.wandb_table_agent.log()

        if output_as_dict:
            return self.metric.result_list_to_dict(result)

        return result

    def _sanity_checks(self) -> None:
        assert self.dataset.dataflow.categories is not None

    def _run_pipe_or_component(self, df_pr: DataFlow) -> DataFlow:
        if self.pipe_component:
            self.pipe_component.put_task(df_pr)
            logger.info("Predicting objects...")
            df_pr_list = self.pipe_component.start()
            if self.wandb_table_agent is not None:
                df_pr_list = [self.wandb_table_agent.dump(dp) for dp in df_pr_list]
            return DataFromList(df_pr_list)
        df_pr = MapData(df_pr, maybe_load_image)
        df_pr = self.pipe.analyze(dataset_dataflow=df_pr, output="image")  # type: ignore
        # deactivate timer for components
        for comp in self.pipe.pipe_component_list:  # type: ignore
            comp.timer_on = False
        df_pr = MapData(df_pr, maybe_remove_image)
        df_list = CacheData(df_pr).get_cache()
        if self.wandb_table_agent is not None:
            df_list = [self.wandb_table_agent.dump(dp) for dp in df_list]
        return DataFromList(df_list)

    def _clean_up_predict_dataflow_annotations(self, df_pr: DataFlow) -> DataFlow:
        # will use the first pipe component of MultiThreadPipelineComponent to get meta annotation
        pipe_or_component = self.pipe_component.pipe_components[0] if self.pipe_component is not None else self.pipe
        meta_anns = pipe_or_component.get_meta_annotation()  # type: ignore
        possible_cats_in_datapoint = self.dataset.dataflow.categories.get_categories(as_dict=False, filtered=True)

        # clean-up procedure depends on the dataset type
        if self.dataset.dataset_info.type == DatasetType.object_detection:
            # we keep all image annotations that will not be generated through processing
            anns_to_keep = {ann for ann in possible_cats_in_datapoint if ann not in meta_anns["image_annotations"]}
            sub_cats_to_remove = meta_anns["sub_categories"]
            relationships_to_remove = meta_anns["relationships"]
            # removing annotations takes place in three steps: First we remove all image annotations. Then, with all
            # remaining image annotations we check, if the image attribute (with Image instance !) is not empty and
            # remove it as well, if necessary. In the last step we remove all sub categories and relationships, if
            # generated in pipeline.
            df_pr = MapData(df_pr, filter_cat(anns_to_keep, possible_cats_in_datapoint))  # pylint: disable=E1120
            df_pr = MapData(df_pr, maybe_remove_image_from_category(anns_to_keep))
            df_pr = MapData(
                df_pr,
                remove_cats(sub_categories=sub_cats_to_remove, relationships=relationships_to_remove),
            )

        elif self.dataset.dataset_info.type == DatasetType.sequence_classification:
            summary_sub_cats_to_remove = meta_anns["summaries"]
            df_pr = MapData(df_pr, remove_cats(summary_sub_categories=summary_sub_cats_to_remove))

        elif self.dataset.dataset_info.type == DatasetType.token_classification:
            sub_cats_to_remove = meta_anns["sub_categories"]
            df_pr = MapData(df_pr, remove_cats(sub_categories=sub_cats_to_remove))
        else:
            raise NotImplementedError

        return df_pr

    def compare(self, interactive: bool = False, **dataflow_build_kwargs: Union[str, int]) -> Optional[ImageType]:
        """
        Visualize ground truth and prediction datapoint. Given a dataflow config it will run predictions per sample
        and concat the prediction image (with predicted bounding boxes) with ground truth image.

        :param interactive: If set to True will open an interactive image, otherwise it will return a numpy array that
                            can be displayed differently (e.g. matplotlib). Note that, if the interactive mode is being
                            used, more than one sample can be iteratively be displayed.
        :param dataflow_build_kwargs: Dataflow configs for displaying specific image splits
        :return: Image as numpy array
        """
        df_gt = self.dataset.dataflow.build(**dataflow_build_kwargs)
        df_pr = self.dataset.dataflow.build(**dataflow_build_kwargs)
        df_gt = MapData(df_gt, maybe_load_image)
        df_pr = MapData(df_pr, maybe_load_image)
        df_pr = MapData(df_pr, deepcopy)
        df_pr = self._clean_up_predict_dataflow_annotations(df_pr)

        page_parsing_component = PageParsingService(
            text_container=LayoutType.word,
            top_level_text_block_names=[
                LayoutType.title,
                LayoutType.text,
                LayoutType.list,
                LayoutType.table,
                LayoutType.figure,
            ],
        )
        df_gt = page_parsing_component.predict_dataflow(df_gt)

        if self.pipe_component:
            pipe_component = self.pipe_component.pipe_components[0]
            df_pr = pipe_component.predict_dataflow(df_pr)
            df_pr = page_parsing_component.predict_dataflow(df_pr)

        elif self.pipe:
            df_pr = self.pipe.analyze(dataset_dataflow=df_pr)
        else:
            raise ValueError("Neither pipe_component nor pipe has been defined")

        df_pr.reset_state()
        df_gt.reset_state()
        for dp_gt, dp_pred in zip(df_gt, df_pr):
            img_gt, img_pred = dp_gt.viz(), dp_pred.viz()
            img_concat = np.concatenate((img_gt, img_pred), axis=1)
            if interactive:
                interactive_imshow(img_concat)
            else:
                return img_concat
        return None


class WandbTableAgent:
    """
    A class that creates a W&B table of sample predictions and sends them to the W&B server.

        df ... # some dataflow
        agent = WandbTableAgent(myrun,"MY_DATASET",50,{"1":"FOO"})
        for dp in df:
            agent.dump(dp)

        agent.log()

    """

    def __init__(
        self,
        wandb_run: "wandb.sdk.wandb_run.Run",
        dataset_name: str,
        num_samples: int,
        categories: Mapping[str, TypeOrStr],
    ):
        """
        :param wandb_run: An `wandb.run` instance for tracking. Use `run=wandb.init(project=project, config=config,
                          **kwargs)` to generate a `run`.
        :param dataset_name: name for tracking
        :param num_samples: When dumping images to a table it will stop adding samples after `num_samples` instances
        :param categories: dict of all possible categories
        """

        self.dataset_name = dataset_name
        self.num_samples = num_samples
        self.categories = categories
        self._run = wandb_run
        self._counter = 0

        # Table logging utils
        self._table_cols: List[str] = ["file_name", "image"]
        self._table_rows: List[Any] = []
        self._table_ref = None

    def dump(self, dp: Image) -> Image:
        """
        Dump image to a table. Add this while iterating over samples. After `num_samples` it will stop appending samples
        to the table

        :param dp: datapoint image
        :return: same as input
        """
        if self.num_samples > self._counter:
            dp = maybe_load_image(dp)
            self._table_rows.append(to_wandb_image(self.categories)(dp))  # pylint: disable=E1102
            dp = maybe_remove_image(dp)
            self._counter += 1
        return dp

    def reset(self) -> None:
        """
        Reset table rows
        """
        self._table_rows = []
        self._counter = 0

    def _build_table(self) -> "Table":
        """
        Builds wandb.Table object for logging evaluation

        returns: Table object to log evaluation
        """
        return Table(columns=self._table_cols, data=self._table_rows)

    def log(self) -> None:
        """
        Log WandB table and maybe send table to WandB server
        """
        table = self._build_table()
        self._run.log({self.dataset_name: table})
        self.reset()

    def _use_table_as_artifact(self) -> None:
        """
        This function logs the given table as artifact and calls `use_artifact`
        on it so tables from next iterations can use the reference of already uploaded images.
        """

        eval_art = Artifact(self._run.id + self.dataset_name, type="dataset")
        eval_art.add(self._build_table(), self.dataset_name)
        self._run.use_artifact(eval_art)
        eval_art.wait()
        self._table_ref = eval_art.get(self.dataset_name).data  # type:ignore



# -*- coding: utf-8 -*-
# File: match.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for matching detections according to various matching rules
"""

from typing import Any, Literal, Optional, Sequence, Tuple, Union

import numpy as np
from numpy.typing import NDArray

from ..datapoint.annotation import ImageAnnotation
from ..datapoint.box import iou
from ..datapoint.image import Image
from ..extern.tp.tpfrcnn.utils.np_box_ops import ioa as np_ioa
from ..utils.settings import TypeOrStr


def match_anns_by_intersection(
    dp: Image,
    parent_ann_category_names: Union[TypeOrStr, Sequence[TypeOrStr]],
    child_ann_category_names: Union[TypeOrStr, Sequence[TypeOrStr]],
    matching_rule: Literal["iou", "ioa"],
    threshold: float,
    use_weighted_intersections: bool = False,
    parent_ann_ids: Optional[Union[Sequence[str], str]] = None,
    child_ann_ids: Optional[Union[str, Sequence[str]]] = None,
    max_parent_only: bool = False,
) -> Tuple[Any, Any, Sequence[ImageAnnotation], Sequence[ImageAnnotation]]:
    """
    Generates an iou/ioa-matrix for parent_ann_categories and child_ann_categories and returns pairs of child/parent
    indices that are above some intersection threshold. It will also return a list of all pre selected parent and child
    annotations.

    **Example:**

    Let `p_i, c_j` be annotations ids of parent and children according to some category names.

    |**ioa**|**c_1**|**c_2**|
    |-------|-------|-------|
    |**p_1**|  0.3  |  0.8  |
    |**p_2**|  0.4  |  0.1  |
    |**p_3**|  1.   |  0.4  |

    With `ioa_threshold = 0.5` it will return:

     `[[2],[0]], [[1],[],[1]], [c_1,c_2], [p_1,p_2,p_3]`.

    For each child the sum of all ioas with all parents sum up to 1. Hence, the ioa with one parent will in general
    decrease if one child intersects with more parents. Take two childs one matching two parents with an ioa of 0.5 each
    while the second matching four parents with an ioa of 0.25 each. In this situation it is difficult to assign
    children according to a given threshold and one also has to take into account the number of parental intersection
    for each child. Setting `use_weighted_intersections` to True will multiply each ioa with the number of intersection
    making it easier to work with an absolute threshold.

    In some situation you want to assign to each child at most one parent. Setting `max_parent_only` to `True` it will
    select the parent with the highest ioa. Note, there is currently no implementation for iou.

    :param dp: image datapoint
    :param parent_ann_category_names: single str or list of category names
    :param child_ann_category_names: single str or list of category names
    :param matching_rule: intersection measure type, either "iou" or "ioa"
    :param threshold: Threshold, for mat given matching rule. Will assign every child ann with iou/ioa above the
                      threshold to the parental annotation.
    :param use_weighted_intersections: This is currently only implemented for matching_rule 'ioa'. Instead of using
                                       the ioa_matrix it will use mat weighted ioa in order to take into account that
                                       intersections with more cells will likely decrease the ioa value. By multiplying
                                       the ioa with the number of all intersection for each child this value calibrate
                                       the ioa.
    :param parent_ann_ids: Additional filter condition. If some ids are selected, it will ignore all other parent candi-
                           dates which are not in the list.
    :param child_ann_ids: Additional filter condition. If some ids are selected, it will ignore all other children
                          candidates which are not in the list.
    :param max_parent_only: Will assign to each child at most one parent with maximum ioa
    :return: child indices, parent indices (see Example), list of parent ids and list of children ids.
    """

    assert matching_rule in ["iou", "ioa"], "matching rule must be either iou or ioa"

    child_anns = dp.get_annotation(annotation_ids=child_ann_ids, category_names=child_ann_category_names)
    child_ann_boxes = np.array(
        [
            ann.image.get_embedding(dp.image_id)
            .transform(dp.width, dp.height, absolute_coords=True)
            .to_list(mode="xyxy")
            for ann in child_anns
            if ann.image is not None
        ]
    )

    # second try, if ann has empty image
    n_dim = child_ann_boxes.ndim
    if n_dim != 2:
        child_ann_boxes = np.array(
            [
                ann.bounding_box.transform(dp.width, dp.height, absolute_coords=True).to_list(mode="xyxy")
                for ann in child_anns
                if ann.bounding_box is not None
            ]
        )

    parent_anns = dp.get_annotation(annotation_ids=parent_ann_ids, category_names=parent_ann_category_names)
    parent_ann_boxes = np.array(
        [
            ann.image.get_embedding(dp.image_id)
            .transform(dp.width, dp.height, absolute_coords=True)
            .to_list(mode="xyxy")
            for ann in parent_anns
            if ann.image is not None
        ]
    )

    # same for parent
    n_dim = parent_ann_boxes.ndim
    if n_dim != 2:
        parent_ann_boxes = np.array(
            [
                ann.bounding_box.transform(dp.width, dp.height, absolute_coords=True).to_list(mode="xyxy")
                for ann in parent_anns
                if ann.bounding_box is not None
            ]
        )

    if matching_rule in ["iou"] and parent_anns and child_anns:
        iou_matrix = iou(child_ann_boxes, parent_ann_boxes)
        output = iou_matrix > threshold
        child_index, parent_index = output.nonzero()
    elif matching_rule in ["ioa"] and parent_anns and child_anns:
        ioa_matrix = np.transpose(np_ioa(parent_ann_boxes, child_ann_boxes))

        if max_parent_only:
            # set all matrix values below threshold to 0
            ioa_matrix[ioa_matrix < threshold] = 0
            # add a dummy column to the left. argmax will choose this column if all ioa values of one child are 0.
            # This index will be ignored in output
            ioa_matrix = np.hstack([np.zeros((ioa_matrix.shape[0], 1)), ioa_matrix])
            child_index_arg_max = ioa_matrix.argmax(1)
            child_index = child_index_arg_max.nonzero()[0]
            child_index_nonzero = child_index_arg_max[child_index]
            # reduce parent index by one, as all indices have been increased by one because of the dummy column
            parent_index = child_index_nonzero - np.ones(child_index_nonzero.shape[0], dtype=np.intc)
        else:

            def _weighted_ioa_matrix(mat: NDArray[np.float32]) -> NDArray[np.float32]:
                sum_of_rows = (mat != 0).sum(1)
                multiplier = np.transpose(sum_of_rows * np.ones((mat.shape[1], mat.shape[0])))
                return multiplier * mat

            if use_weighted_intersections:
                ioa_matrix = _weighted_ioa_matrix(ioa_matrix)
            output = ioa_matrix > threshold
            child_index, parent_index = output.nonzero()

    else:
        return [], [], [], []

    return child_index, parent_index, child_anns, parent_anns



# -*- coding: utf-8 -*-
# File: test_pdftext.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module extern.pdftext
"""
from pytest import mark

from deepdoctection.extern.pdftext import PdfPlumberTextDetector


class TestPdfPlumberTextDetector:
    """
    Test PdfPlumberTextDetector
    """

    @staticmethod
    @mark.requires_tf_or_pt
    def test_pdf_plumber_detector_predicts_image(pdf_bytes: bytes) -> None:
        """
        PdfPlumber returns words from pdf_bytes correctly
        """

        # Arrange
        pdf_text_predictor = PdfPlumberTextDetector()

        # Act
        detect_results_list = pdf_text_predictor.predict(pdf_bytes)

        # Assert
        assert len(detect_results_list) == 109
        assert detect_results_list[0].text == "A"
        assert detect_results_list[1].text == "Simple"

    @staticmethod
    @mark.requires_tf_or_pt
    def test_pdf_plumber_detector_returns_width_height(pdf_bytes: bytes, pdf_bytes_page_2: bytes) -> None:
        """
        PdfPlumber returns pdf width and height correctly
        """

        # Arrange
        pdf_text_predictor = PdfPlumberTextDetector()
        _ = pdf_text_predictor.predict(pdf_bytes)

        # Act
        width, height = pdf_text_predictor.get_width_height(pdf_bytes)

        # Assert
        assert width == 612
        assert height == 792

        # Act
        width, height = pdf_text_predictor.get_width_height(pdf_bytes_page_2)
        assert width == 1112
        assert height == 1792



# -*- coding: utf-8 -*-
# File: custom.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Adding some functionality to dataflow classes (e.g. monkey patching, inheritance ...). Some ideas have been taken
from

<https://github.com/tensorpack/dataflow/blob/master/dataflow/dataflow/common.py>
"""
from typing import Any, Callable, Iterable, Iterator, List, Optional

import numpy as np

from ..utils.logger import logger
from ..utils.tqdm import get_tqdm
from ..utils.utils import get_rng
from .base import DataFlow, DataFlowReentrantGuard, DataFlowResetStateNotCalled, ProxyDataFlow
from .serialize import DataFromIterable, DataFromList

__all__ = ["CacheData", "CustomDataFromList", "CustomDataFromIterable"]


class CacheData(ProxyDataFlow):
    """
    Completely cache the first pass of a DataFlow in memory,
    and produce from the cache thereafter.

    NOTE: The user should not stop the iterator before it has reached the end.
    Otherwise, the cache may be incomplete.

    **Example:**

            df_list = CacheData(df).get_cache()   # buffers the whole dataflow and return a list of all datapoints
    """

    def __init__(self, df: DataFlow, shuffle: bool = False) -> None:
        """
        :param df: input DataFlow.
        :param shuffle: whether to shuffle the cache before yielding from it.
        """
        self.shuffle = shuffle
        self.buffer: List[Any] = []
        self._guard: Optional[DataFlowReentrantGuard] = None
        self.rng = get_rng(self)
        super().__init__(df)

    def reset_state(self) -> None:
        super().reset_state()
        self._guard = DataFlowReentrantGuard()
        self.buffer = []

    def __iter__(self) -> Iterator[Any]:
        if self._guard is None:
            raise DataFlowResetStateNotCalled()

        with self._guard:
            if self.buffer:
                if self.shuffle:
                    self.rng.shuffle(self.buffer)
                yield from self.buffer
            else:
                for dp in self.df:
                    yield dp
                    self.buffer.append(dp)

    def get_cache(self) -> List[Any]:
        """
        get the cache of the whole dataflow as a list

        :return: list of datapoints
        """
        self.reset_state()
        with get_tqdm() as status_bar:
            for _ in self:
                status_bar.update()  # type: ignore
            if self.shuffle:
                self.rng.shuffle(self.buffer)
            return self.buffer


class CustomDataFromList(DataFromList):
    """
    Wraps a list of datapoints to a dataflow. Compared to `Tensorpack.DataFlow.DataFromList` implementation you
    can specify a number of datapoints after that the iteration stops. You can also pass a rebalance function that
    filters on that list.

    **Example:**

                def filter_first(lst):
                    return lst.pop(0)

                df = CustomDataFromList(lst=[["a","b"],["c","d"]],rebalance_func=filter_first)
                df.reset_state()

        will yield:

            ["c","d"]

    """

    def __init__(
        self,
        lst: List[Any],
        shuffle: bool = False,
        max_datapoints: Optional[int] = None,
        rebalance_func: Optional[Callable[[List[Any]], List[Any]]] = None,
    ):
        """
        :param lst: the input list. Each element represents a datapoint.
        :param shuffle: Whether to shuffle the list before streaming.
        :param max_datapoints: The maximum number of datapoints to return before stopping the iteration.
                               If None it streams the whole dataflow.
        :param rebalance_func: A func that inputs a list and outputs a list. Useful, if you want to filter the passed
                               list and re-balance the sample. Only the output list of the re-balancing function will be
                               considered.
        """
        super().__init__(lst, shuffle)
        self.max_datapoints = max_datapoints
        self.rebalance_func = rebalance_func

    def __len__(self) -> int:
        if self.max_datapoints is not None:
            return min(self.max_datapoints, len(self.lst))
        return len(self.lst)

    def __iter__(self) -> Iterator[Any]:
        if self.rng is None:
            raise DataFlowResetStateNotCalled()
        if self.rebalance_func is not None:
            lst_tmp = self.rebalance_func(self.lst)
            logger.info("subset size after re-balancing: %s", len(lst_tmp))
        else:
            lst_tmp = self.lst

        if self.shuffle:
            idxs = np.arange(len(lst_tmp))
            self.rng.shuffle(idxs)
            for idx, k in enumerate(idxs):
                if self.max_datapoints is not None:
                    if idx < self.max_datapoints:
                        yield lst_tmp[k]
                    else:
                        break
                else:
                    yield lst_tmp[k]
        else:
            if self.max_datapoints is not None:
                for k, _ in enumerate(lst_tmp):
                    if k < self.max_datapoints:
                        yield lst_tmp[k]
                    else:
                        break
            else:
                yield from lst_tmp


class CustomDataFromIterable(DataFromIterable):
    """
    Wrap an iterable of datapoints to a dataflow. Can stop iteration after max_datapoints.
    """

    def __init__(self, iterable: Iterable[Any], max_datapoints: Optional[int] = None):
        """
        :param iterable: An iterable object
        :param max_datapoints: The maximum number of datapoints to stream. If None it iterates through the whole
                               dataflow.
        """
        super().__init__(iterable)
        self.max_datapoints = max_datapoints
        if self.max_datapoints is not None:
            self._len = self.max_datapoints

    def __iter__(self) -> Any:
        if self.max_datapoints is not None:
            for idx, dp in enumerate(self._itr):
                if idx < self.max_datapoints:
                    yield dp
                else:
                    break
        else:
            yield from self._itr



# -*- coding: utf-8 -*-
# File: test_d2struct.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module mapper.d2struct
"""
from unittest.mock import MagicMock, patch

from pytest import mark

from deepdoctection.datapoint.image import Image
from deepdoctection.utils import detectron2_available

from .data import DatapointImage

if detectron2_available():
    from deepdoctection.mapper.d2struct import image_to_d2_frcnn_training


@mark.requires_pt
@patch("deepdoctection.mapper.d2struct.os.path.isfile", MagicMock(return_value=True))
def test_image_to_d2_frcnn_training(datapoint_image: Image, image_results: DatapointImage) -> None:
    """
    testing image_to_d2_frcnn_training is mapping correctly
    """

    # Act
    img_to_d2_tr_mapper = image_to_d2_frcnn_training(add_mask=False)  # pylint: disable=E1120
    output = img_to_d2_tr_mapper(datapoint_image)

    # Assert
    expected_output = image_results.get_d2_frcnn_training_anns()

    assert output
    assert output.keys() == expected_output.keys()
    assert output["width"] == output["width"]

    first_annotation = output["annotations"][0]
    expected_first_annotation = expected_output["annotations"][0]

    assert first_annotation["bbox"] == expected_first_annotation["bbox"]
    assert first_annotation["category_id"] == expected_first_annotation["category_id"]



# -*- coding: utf-8 -*-
# File: tools.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for tools around processing documents with Pipelines.
"""

import json
import os

from ..dataflow.base import DataFlow
from ..dataflow.common import MapData
from ..dataflow.custom_serialize import SerializerJsonlines
from ..datapoint.view import Page
from ..utils.detection_types import Pathlike
from ..utils.utils import is_file_extension


def load_page(path: Pathlike) -> Page:
    """
    Load a `json` file and generate a page object.

    :param path: Path to load from
    :return: A page object
    """

    assert os.path.isfile(path), path
    file_name = os.path.split(path)[1]
    assert is_file_extension(file_name, ".json"), file_name

    with open(path, "r", encoding="UTF-8") as file:
        page_dict = json.load(file)

    return Page.from_dict(**page_dict)


def load_document(path: Pathlike) -> DataFlow:
    """
    Load a parsed document from a `jsonl` file and generate a DataFlow that can be streamed.

    :param path: Path to a `jsonl` file
    :return: DataFlow
    """
    df: DataFlow
    df = SerializerJsonlines.load(path)
    df = MapData(df, Page.from_dict)
    return df



# -*- coding: utf-8 -*-
# File: transform.py

# Copyright 2022 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for transform style pipeline components. These pipeline components are used for various transforming operations
on images (e.g. deskew, de-noising or more general GAN like operations.
"""

from ..datapoint.image import Image
from ..extern.base import ImageTransformer
from ..utils.detection_types import JsonDict
from ..utils.logger import logger
from .base import ImageTransformPipelineComponent
from .registry import pipeline_component_registry


@pipeline_component_registry.register("SimpleTransformService")
class SimpleTransformService(ImageTransformPipelineComponent):
    """
    Pipeline component for transforming an image. The service is designed for applying transform predictors that
    take an image as numpy array as input and return the same. The service itself will change the underlying metadata
    like height and width of the returned transform.

    This component is meant to be used at the very first stage of a pipeline. If components have already returned image
    annotations then this component will currently not re-calculate bounding boxes in terms of the transformed image.
    It will raise a warning (at runtime) if image annotations have already been appended.
    """

    def __init__(self, transform_predictor: ImageTransformer):
        """

        :param transform_predictor: image transformer
        """
        super().__init__(self._get_name(transform_predictor.name), transform_predictor)

    def serve(self, dp: Image) -> None:
        if dp.annotations:
            logger.warning(
                "%s has already received image with image annotations. These annotations will not "
                "be transformed and might cause unexpected output in your pipeline.",
                self.name,
            )
        if dp.image is not None:
            np_image_transform = self.transform_predictor.transform(dp.image)
            self.dp_manager.datapoint.clear_image(True)
            self.dp_manager.datapoint.image = np_image_transform

    def clone(self) -> "SimpleTransformService":
        return self.__class__(self.transform_predictor)

    def get_meta_annotation(self) -> JsonDict:
        return dict(
            [
                ("image_annotations", []),
                ("sub_categories", {}),
                ("relationships", {}),
                ("summaries", []),
            ]
        )

    @staticmethod
    def _get_name(transform_name: str) -> str:
        return f"simple_transform_{transform_name}"



# -*- coding: utf-8 -*-
# File: hfdetr.py

# Copyright 2022 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
HF Detr model for object detection.
"""

from typing import List, Literal, Mapping, Optional, Sequence

from ..utils.detection_types import ImageType, Requirement
from ..utils.file_utils import (
    get_pytorch_requirement,
    get_transformers_requirement,
    pytorch_available,
    transformers_available,
)
from ..utils.settings import TypeOrStr, get_type
from .base import DetectionResult, ObjectDetector
from .pt.ptutils import set_torch_auto_device

if pytorch_available():
    import torch  # pylint: disable=W0611
    from torchvision.ops import boxes as box_ops  # type: ignore

if transformers_available():
    from transformers import (  # pylint: disable=W0611
        AutoFeatureExtractor,
        DetrFeatureExtractor,
        PretrainedConfig,
        TableTransformerForObjectDetection,
    )


def _detr_post_processing(
    boxes: "torch.Tensor", scores: "torch.Tensor", labels: "torch.Tensor", nms_thresh: float
) -> "torch.Tensor":
    return box_ops.batched_nms(boxes.float(), scores, labels, nms_thresh)


def detr_predict_image(
    np_img: ImageType,
    predictor: "TableTransformerForObjectDetection",
    feature_extractor: "DetrFeatureExtractor",
    device: Literal["cpu", "cuda"],
    threshold: float,
    nms_threshold: float,
) -> List[DetectionResult]:
    """
    Calling predictor. Before doing that, tensors must be transferred to the device where the model is loaded. After
    prediction it will present prediction in DetectionResult format-

    :param np_img: image as numpy array
    :param predictor: TableTransformerForObjectDetection
    :param feature_extractor: feature extractor
    :param device: device where the model is loaded
    :param threshold: Will filter all predictions with confidence score less threshold
    :param nms_threshold: Threshold to perform NMS on prediction outputs. (Note, that NMS does not belong to canonical
                          Detr inference processing)

    """
    target_sizes = [np_img.shape[:2]]
    inputs = feature_extractor(images=np_img, return_tensors="pt")
    inputs.data["pixel_values"] = inputs.data["pixel_values"].to(device)
    inputs.data["pixel_mask"] = inputs.data["pixel_mask"].to(device)
    outputs = predictor(**inputs)
    outputs["encoder_last_hidden_state"] = outputs["encoder_last_hidden_state"].to("cpu")
    outputs["last_hidden_state"] = outputs["last_hidden_state"].to("cpu")
    outputs["logits"] = outputs["logits"].to("cpu")
    outputs["pred_boxes"] = outputs["pred_boxes"].to("cpu")
    results = feature_extractor.post_process_object_detection(outputs, threshold=threshold, target_sizes=target_sizes)[
        0
    ]
    keep = _detr_post_processing(results["boxes"], results["scores"], results["labels"], nms_threshold)
    keep_boxes = results["boxes"][keep]
    keep_scores = results["scores"][keep]
    keep_labels = results["labels"][keep]
    return [
        DetectionResult(box=box.tolist(), score=score.item(), class_id=class_id.item())
        for box, score, class_id in zip(keep_boxes, keep_scores, keep_labels)
    ]


class HFDetrDerivedDetector(ObjectDetector):
    """
    Model wrapper for TableTransformerForObjectDetection that again is based on

    https://github.com/microsoft/table-transformer .

    The wrapper can be used to load pre-trained models for table detection and table structure recognition. Running Detr
    models trained from scratch on custom datasets is possible as well. Note, that this wrapper will load
    `TableTransformerForObjectDetection` that is slightly different compared to `DetrForObjectDetection` that can be
    found in the transformer library as well.

        config_path = ModelCatalog.
        get_full_path_configs("microsoft/table-transformer-structure-recognition/pytorch_model.bin")
        weights_path = ModelDownloadManager.
        get_full_path_weights("microsoft/table-transformer-structure-recognition/pytorch_model.bin")
        feature_extractor_config_path = ModelDownloadManager.
        get_full_path_preprocessor_configs("microsoft/table-transformer-structure-recognition/pytorch_model.bin")
        categories = ModelCatalog.
        get_profile("microsoft/table-transformer-structure-recognition/pytorch_model.bin").categories

        detr_predictor = HFDetrDerivedDetector(config_path,weights_path,feature_extractor_config_path,categories)

        detection_result = detr_predictor.predict(bgr_image_np_array)
    """

    def __init__(
        self,
        path_config_json: str,
        path_weights: str,
        path_feature_extractor_config_json: str,
        categories: Mapping[str, TypeOrStr],
        device: Optional[Literal["cpu", "cuda"]] = None,
        filter_categories: Optional[Sequence[TypeOrStr]] = None,
    ):
        """
        Set up the predictor.
        :param path_config_json: The path to the json config.
        :param path_weights: The path to the model checkpoint.
        :param path_feature_extractor_config_json: The path to the feature extractor config.
        :param categories: A dict with key (indices) and values (category names).
        :param device: "cpu" or "cuda". If not specified will auto select depending on what is available
        :param filter_categories: The model might return objects that are not supposed to be predicted and that should
                                  be filtered. Pass a list of category names that must not be returned
        """
        self.name = "Detr"
        self.categories = {idx: get_type(cat) for idx, cat in categories.items()}
        self.path_config = path_config_json
        self.path_weights = path_weights
        self.path_feature_extractor_config = path_feature_extractor_config_json
        self.config = PretrainedConfig.from_pretrained(pretrained_model_name_or_path=self.path_config)
        self.config.threshold = 0.1
        self.config.nms_threshold = 0.05
        self.hf_detr_predictor = self.set_model(path_weights)
        self.feature_extractor = self.set_pre_processor()

        if device is not None:
            self.device = device
        else:
            self.device = set_torch_auto_device()
        self.hf_detr_predictor.to(self.device)
        if filter_categories:
            filter_categories = [get_type(cat) for cat in filter_categories]
        self.filter_categories = filter_categories

    def predict(self, np_img: ImageType) -> List[DetectionResult]:
        results = detr_predict_image(
            np_img,
            self.hf_detr_predictor,
            self.feature_extractor,
            self.device,
            self.config.threshold,
            self.config.nms_threshold,
        )
        return self._map_category_names(results)

    def set_model(self, path_weights: str) -> "TableTransformerForObjectDetection":
        """
        Builds the Detr model

        :param path_weights: weights
        :return: TableTransformerForObjectDetection instance
        """
        return TableTransformerForObjectDetection.from_pretrained(
            pretrained_model_name_or_path=path_weights, config=self.config
        )

    def set_pre_processor(self) -> "DetrFeatureExtractor":
        """
        Builds the feature extractor

        :return: DetrFeatureExtractor
        """
        return AutoFeatureExtractor.from_pretrained(pretrained_model_name_or_path=self.path_feature_extractor_config)

    def _map_category_names(self, detection_results: List[DetectionResult]) -> List[DetectionResult]:
        """
        Populating category names to detection results. Will also filter categories

        :param detection_results: list of detection results
        :return: List of detection results with attribute class_name populated
        """
        filtered_detection_result: List[DetectionResult] = []
        for result in detection_results:
            result.class_name = self.categories[str(result.class_id + 1)]  # type: ignore
            if isinstance(result.class_id, int):
                result.class_id += 1
            if self.filter_categories:
                if result.class_name not in self.filter_categories:
                    filtered_detection_result.append(result)
            else:
                filtered_detection_result.append(result)

        return filtered_detection_result

    @classmethod
    def get_requirements(cls) -> List[Requirement]:
        return [get_pytorch_requirement(), get_transformers_requirement()]

    def clone(self) -> "HFDetrDerivedDetector":
        return self.__class__(
            self.path_config, self.path_weights, self.path_feature_extractor_config, self.categories, self.device
        )



# -*- coding: utf-8 -*-
# File: test_laylmstruct.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing the module mapper.laylmstruct
"""


from typing import Dict
from unittest.mock import MagicMock, patch

import numpy as np
from pytest import mark

from deepdoctection.datapoint import Image
from deepdoctection.mapper.laylmstruct import image_to_raw_layoutlm_features, raw_features_to_layoutlm_features
from deepdoctection.mapper.xfundstruct import xfund_to_image
from deepdoctection.utils.detection_types import JsonDict
from deepdoctection.utils.file_utils import transformers_available
from deepdoctection.utils.settings import DatasetType, ObjectTypes, WordType

if transformers_available():
    from transformers import LayoutLMTokenizerFast


@mark.basic
@patch("deepdoctection.mapper.xfundstruct.load_image_from_file", MagicMock(return_value=np.ones((1000, 1000, 3))))
def test_image_to_raw_layoutlm_features_for_token_data(
    datapoint_xfund: JsonDict,
    xfund_category_names: Dict[str, str],
    xfund_category_dict: Dict[ObjectTypes, str],
    raw_layoutlm_features: JsonDict,
    ner_token_to_id_mapping: JsonDict,
) -> None:
    """
    testing image_to_raw_layoutlm_features is mapping correctly for dataset type "TOKEN_CLASSIFICATION"
    """

    # Arrange
    image = xfund_to_image(True, False, xfund_category_dict, xfund_category_names, ner_token_to_id_mapping)(
        datapoint_xfund
    )

    # Act
    raw_features = image_to_raw_layoutlm_features(DatasetType.token_classification)(image)

    # Assert
    assert raw_features is not None
    assert raw_features["image_id"] == raw_layoutlm_features["image_id"]
    assert raw_features["width"] == raw_layoutlm_features["width"]
    assert raw_features["height"] == raw_layoutlm_features["height"]
    assert raw_features["ann_ids"] == raw_layoutlm_features["ann_ids"]
    assert raw_features["words"] == raw_layoutlm_features["words"]
    assert raw_features["bbox"] == raw_layoutlm_features["bbox"]
    assert raw_features["dataset_type"] == raw_layoutlm_features["dataset_type"]
    assert raw_features["labels"] == raw_layoutlm_features["labels"]


@mark.basic
@patch("deepdoctection.mapper.xfundstruct.load_image_from_file", MagicMock(return_value=np.ones((1000, 1000, 3))))
def test_image_to_raw_layoutlm_features_for_inference(
    datapoint_xfund: JsonDict,
    xfund_category_names: Dict[str, str],
    xfund_category_dict: Dict[ObjectTypes, str],
    raw_layoutlm_features: JsonDict,
    ner_token_to_id_mapping: JsonDict,
) -> None:
    """
    testing image_to_raw_layoutlm_features is mapping correctly. Semantic entities and tags have been removed, so that
    the scenario covers the inference case
    """

    # Arrange
    image = xfund_to_image(True, False, xfund_category_dict, xfund_category_names, ner_token_to_id_mapping)(
        datapoint_xfund
    )

    assert image is not None

    for ann in image.get_annotation():
        ann.remove_sub_category(WordType.token_class)
        ann.remove_sub_category(WordType.tag)

    # Act
    raw_features = image_to_raw_layoutlm_features(DatasetType.token_classification)(image)

    # Assert
    assert raw_features is not None
    assert raw_features["image_id"] == raw_layoutlm_features["image_id"]
    assert raw_features["width"] == raw_layoutlm_features["width"]
    assert raw_features["height"] == raw_layoutlm_features["height"]
    assert raw_features["ann_ids"] == raw_layoutlm_features["ann_ids"]
    assert raw_features["words"] == raw_layoutlm_features["words"]
    assert raw_features["bbox"] == raw_layoutlm_features["bbox"]
    assert raw_features["dataset_type"] == raw_layoutlm_features["dataset_type"]
    assert "labels" not in raw_features


@mark.requires_pt
def test_raw_features_to_layoutlm_features(
    dp_image_with_layout_and_word_annotations: Image, layoutlm_features: JsonDict
) -> None:
    """
    testing image_to_layoutlm_features is mapping correctly
    """

    # Arrange
    dp = image_to_raw_layoutlm_features()(dp_image_with_layout_and_word_annotations)  # pylint: disable=E1120
    tokenizer_fast = LayoutLMTokenizerFast.from_pretrained("microsoft/layoutlm-base-uncased")

    # Act
    assert dp is not None
    feature = raw_features_to_layoutlm_features(dp, tokenizer_fast, padding="do_not_pad")

    # Assert
    # pylint: disable=E1136
    assert feature["image_ids"] == layoutlm_features["image_ids"]
    assert feature["width"] == layoutlm_features["width"]
    assert feature["height"] == layoutlm_features["height"]
    assert feature["ann_ids"] == layoutlm_features["ann_ids"]
    assert feature["input_ids"] == layoutlm_features["input_ids"]
    assert feature["token_type_ids"] == layoutlm_features["token_type_ids"]
    assert feature["attention_mask"] == layoutlm_features["attention_mask"]
    assert feature["bbox"] == layoutlm_features["bbox"]
    assert feature["tokens"] == layoutlm_features["tokens"]
    # pylint: enable=E1136



# -*- coding: utf-8 -*-
# File: doctrocr.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Deepdoctection wrappers for DocTr OCR text line detection and text recognition models
"""

from typing import List, Tuple

from ..utils.detection_types import ImageType, Requirement
from ..utils.file_utils import (
    doctr_available,
    get_doctr_requirement,
    get_pytorch_requirement,
    get_tensorflow_requirement,
    get_tf_addons_requirements,
    pytorch_available,
    tf_addons_available,
    tf_available,
)
from ..utils.settings import LayoutType, ObjectTypes
from .base import DetectionResult, ObjectDetector, PredictorBase, TextRecognizer

if doctr_available() and ((tf_addons_available() and tf_available()) or pytorch_available()):
    from doctr.models.detection.predictor import DetectionPredictor  # pylint: disable=W0611
    from doctr.models.detection.zoo import detection_predictor
    from doctr.models.recognition.predictor import RecognitionPredictor  # pylint: disable=W0611
    from doctr.models.recognition.zoo import recognition_predictor


def doctr_predict_text_lines(np_img: ImageType, predictor: "DetectionPredictor") -> List[DetectionResult]:
    """
    Generating text line DetectionResult based on Doctr DetectionPredictor.

    :param np_img: Image in np.array.
    :param predictor: `doctr.models.detection.predictor.DetectionPredictor`
    :return: A list of text line detection results (without text).
    """
    raw_output = predictor([np_img])
    detection_results = [
        DetectionResult(
            box=box[:4].tolist(), class_id=1, score=box[4], absolute_coords=False, class_name=LayoutType.word
        )
        for box in raw_output[0]
    ]
    return detection_results


def doctr_predict_text(inputs: List[Tuple[str, ImageType]], predictor: "RecognitionPredictor") -> List[DetectionResult]:
    """
    Calls Doctr text recognition model on a batch of numpy arrays (text lines predicted from a text line detector) and
    returns the recognized text as DetectionResult

    :param inputs: list of tuples containing the annotation_id of the input image and the numpy array of the cropped
                   text line

    :param predictor: `doctr.models.detection.predictor.RecognitionPredictor`
    :return: A list of DetectionResult containing recognized text.
    """

    uuids, images = list(zip(*inputs))
    raw_output = predictor(list(images))
    detection_results = [
        DetectionResult(score=output[1], text=output[0], uuid=uuid) for uuid, output in zip(uuids, raw_output)
    ]
    return detection_results


class DoctrTextlineDetector(ObjectDetector):
    """
    A deepdoctection wrapper of DocTr text line detector. We model text line detection as ObjectDetector
    and assume to use this detector in a ImageLayoutService. This model currently uses the default implementation
    DBNet as described in “Real-time Scene Text Detection with Differentiable Binarization”, using a ResNet-50 backbone.
    and can be used in either Tensorflow or PyTorch.

    Regarding the model we refer to the documentation <https://mindee.github.io/doctr/models.html>

    **Example:**

                 path = "/path/to/image_dir"
                 det = DoctrTextlineDetector()
                 layout = ImageLayoutService(det,to_image=True, crop_image=True)
                 rec = DoctrTextRecognizer()
                 text = TextExtractionService(rec,extract_from_roi="WORD")
                 analyzer = DoctectionPipe(pipeline_component_list=[layout,text])
                 df = analyzer.analyze(path = path)

                 for dp in df:
                     ...


    """

    def __init__(self) -> None:
        self.name = "doctr_text_detector"
        self.doctr_predictor = detection_predictor(pretrained=True)
        self.categories = {"1": LayoutType.word}

    def predict(self, np_img: ImageType) -> List[DetectionResult]:
        """
        Prediction per image.

        :param np_img: image as numpy array
        :return: A list of DetectionResult
        """
        detection_results = doctr_predict_text_lines(np_img, self.doctr_predictor)
        return detection_results

    @classmethod
    def get_requirements(cls) -> List[Requirement]:
        if tf_available():
            return [get_tensorflow_requirement(), get_doctr_requirement(), get_tf_addons_requirements()]
        if pytorch_available():
            return [get_pytorch_requirement(), get_doctr_requirement()]
        raise ModuleNotFoundError("Neither Tensorflow nor PyTorch has been installed. Cannot use DoctrTextlineDetector")

    def clone(self) -> PredictorBase:
        return self.__class__()

    def possible_categories(self) -> List[ObjectTypes]:
        return [LayoutType.word]


class DoctrTextRecognizer(TextRecognizer):
    """
    A deepdoctection wrapper of DocTr text recognition predictor. The base class is a TextRecognizer that takes
    a batch of sub images (e.g. text lines from a text detector) and returns a list with text spotted in the sub images.
    This model currently uses the default implementation CRNN with a VGG-16 backbone as described in
    “An End-to-End Trainable Neural Network for Image-based Sequence Recognition and Its Application to Scene
    Text Recognition” and can be used in either Tensorflow or PyTorch.

    Regarding the model we refer to the documentation <https://mindee.github.io/doctr/models.html>

    **Example:**

                 path = "/path/to/image_dir"
                 det = DoctrTextlineDetector()
                 layout = ImageLayoutService(det,to_image=True, crop_image=True)
                 rec = DoctrTextRecognizer()
                 text = TextExtractionService(rec,extract_from_roi="LINE")
                 analyzer = DoctectionPipe(pipeline_component_list=[layout,text])
                 df = analyzer.analyze(path = path)

                 for dp in df:
                     ...

    """

    def __init__(self) -> None:
        self.name = "doctr_text_recognizer"
        self.doctr_predictor = recognition_predictor(pretrained=True)

    def predict(self, images: List[Tuple[str, ImageType]]) -> List[DetectionResult]:
        """
        Prediction on a batch of text lines

        :param images: list of tuples with the annotation_id of the sub image and a numpy array
        :return: A list of DetectionResult
        """
        if images:
            return doctr_predict_text(images, self.doctr_predictor)
        return []

    @classmethod
    def get_requirements(cls) -> List[Requirement]:
        if tf_available():
            return [get_tensorflow_requirement(), get_doctr_requirement(), get_tf_addons_requirements()]
        if pytorch_available():
            return [get_pytorch_requirement(), get_doctr_requirement()]
        raise ModuleNotFoundError("Neither Tensorflow nor PyTorch has been installed. Cannot use DoctrTextRecognizer")

    def clone(self) -> PredictorBase:
        return self.__class__()



# -*- coding: utf-8 -*-
# File: tp_frcnn_train.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for training Tensorpack `GeneralizedRCNN`
"""

import os
from typing import Dict, List, Optional, Sequence, Type, Union

# pylint: disable=import-error
from tensorpack.callbacks import (
    EstimatedTimeLeft,
    GPUMemoryTracker,
    GPUUtilizationTracker,
    HostMemoryTracker,
    ModelSaver,
    PeriodicCallback,
    ScheduledHyperParamSetter,
    SessionRunTimeout,
    ThroughputTracker,
)

# todo: check how dataflow import is directly possible without having AssertionError
from tensorpack.dataflow import ProxyDataFlow, imgaug
from tensorpack.input_source import QueueInput
from tensorpack.tfutils import SmartInit
from tensorpack.train import SyncMultiGPUTrainerReplicated, TrainConfig, launch_train_with_config
from tensorpack.utils import logger

from ..dataflow.base import DataFlow
from ..dataflow.common import MapData
from ..dataflow.parallel_map import MultiProcessMapData
from ..dataflow.serialize import DataFromList
from ..datasets.base import DatasetBase
from ..eval.base import MetricBase
from ..eval.registry import metric_registry
from ..eval.tp_eval_callback import EvalCallback
from ..extern.tp.tfutils import disable_tfv2
from ..extern.tp.tpfrcnn.common import CustomResize
from ..extern.tp.tpfrcnn.config.config import model_frcnn_config, train_frcnn_config
from ..extern.tp.tpfrcnn.modeling.generalized_rcnn import ResNetFPNModel
from ..extern.tp.tpfrcnn.preproc import anchors_and_labels, augment
from ..extern.tpdetect import TPFrcnnDetector
from ..mapper.maputils import LabelSummarizer
from ..mapper.tpstruct import image_to_tp_frcnn_training
from ..pipe.base import PredictorPipelineComponent
from ..pipe.registry import pipeline_component_registry
from ..utils.detection_types import JsonDict
from ..utils.file_utils import set_mp_spawn
from ..utils.fs import get_load_image_func
from ..utils.logger import log_once
from ..utils.metacfg import AttrDict, set_config_by_yaml
from ..utils.tqdm import get_tqdm
from ..utils.utils import string_to_dict

__all__ = ["train_faster_rcnn"]


class LoadAugmentAddAnchors:
    """
    A helper class for default mapping `load_augment_add_anchors`.
    """

    def __init__(self, config: AttrDict) -> None:
        self.cfg = config

    def __call__(self, dp: JsonDict) -> Optional[JsonDict]:
        return load_augment_add_anchors(dp, self.cfg)


def load_augment_add_anchors(dp: JsonDict, config: AttrDict) -> Optional[JsonDict]:
    """
    Transforming an image before entering the graph. This function bundles all the necessary steps to feed
    the network for training.

    :param dp: A dict with 'file_name', 'gt_boxes', 'gt_labels' and optional 'image'
    :param config: An `AttrDict` with a TP frcnn config
    :return: An dict with all necessary keys for feeding the graph
    """
    cfg = config
    if "image" not in dp:
        loader = get_load_image_func(dp["file_name"])
        dp["image"] = loader(dp["file_name"])

    augment_list = [
        CustomResize(cfg.PREPROC.TRAIN_SHORT_EDGE_SIZE, cfg.PREPROC.MAX_SIZE),
        imgaug.Flip(horiz=True),
    ]
    dp = augment(dp, augment_list, False)
    dp_with_anchors = anchors_and_labels(
        dp,
        cfg.FPN.ANCHOR_STRIDES,
        cfg.RPN.ANCHOR_SIZES,
        cfg.RPN.ANCHOR_RATIOS,
        cfg.PREPROC.MAX_SIZE,
        cfg.FRCNN.BATCH_PER_IM,
        cfg.RPN.FG_RATIO,
        cfg.RPN.POSITIVE_ANCHOR_THRESH,
        cfg.RPN.NEGATIVE_ANCHOR_THRESH,
        cfg.RPN.CROWD_OVERLAP_THRESH,
    )
    if dp_with_anchors is not None:
        dp_with_anchors.pop("file_name")
    return dp_with_anchors


def get_train_dataflow(
    dataset: DatasetBase, config: AttrDict, use_multi_proc_for_train: bool, **build_train_kwargs: str
) -> DataFlow:
    """
    Return a dataflow for training TP Frcnn. The returned dataflow depends on the dataset and the configuration of
    the model, as the augmentation is part of the data preparation.

    :param dataset: A dataset for object detection
    :param config: An `AttrDict` with a TP Frcnn config
    :param use_multi_proc_for_train: If set to `True` will use multi processes for augmenting
    :param build_train_kwargs: build configuration of the dataflow.
    :return: A dataflow
    """

    set_mp_spawn()
    cfg = config
    df = dataset.dataflow.build(**build_train_kwargs)
    df = MapData(df, image_to_tp_frcnn_training(add_mask=False))  # pylint: disable=E1120

    logger.info("Loading dataset into memory")

    max_datapoints: Optional[int] = int(build_train_kwargs.get("max_datapoints", 0))
    if not max_datapoints:
        max_datapoints = None

    datapoints = []
    summarizer = LabelSummarizer(cfg.DATA.CLASS_DICT)
    df.reset_state()

    with get_tqdm(total=max_datapoints) as status_bar:
        for dp in df:
            if "image" in dp:
                log_once(
                    "Datapoint have images as np arrays stored and they will be loaded into memory. "
                    "To avoid OOM set 'load_image'=False in dataflow build config. This will load "
                    "images when needed and reduce memory costs!!!",
                    "warn",
                )
            summarizer.dump(dp["gt_labels"])
            datapoints.append(dp)
            status_bar.update()  # type: ignore
    summarizer.print_summary_histogram(dd_logic=False)
    num_datapoints = len(datapoints)
    logger.info("Total #images for training: %i", num_datapoints)
    df = DataFromList(datapoints, shuffle=True)
    buffer_size = min(num_datapoints - 1, 200)

    load_augment_anchors = LoadAugmentAddAnchors(cfg)  # can't use dec: curry as pickling will fail in mp
    if use_multi_proc_for_train:
        num_cpu = os.cpu_count()
        if num_cpu is None:
            num_cpu = 0
        df = MultiProcessMapData(
            df,
            num_proc=1 if buffer_size < 3 else num_cpu // 2,
            map_func=load_augment_anchors,
            buffer_size=buffer_size,
        )
    else:
        df = MapData(df, load_augment_anchors)
    return ProxyDataFlow(df)


def train_faster_rcnn(
    path_config_yaml: str,
    dataset_train: DatasetBase,
    path_weights: str = "",
    config_overwrite: Optional[List[str]] = None,
    log_dir: str = "train_log/frcnn",
    build_train_config: Optional[Sequence[str]] = None,
    dataset_val: Optional[DatasetBase] = None,
    build_val_config: Optional[Sequence[str]] = None,
    metric_name: Optional[str] = None,
    metric: Optional[Union[Type[MetricBase], MetricBase]] = None,
    pipeline_component_name: Optional[str] = None,
) -> None:
    """
    Easy adaptation of the training script for Tensorpack Faster-RCNN.

    Train Faster-RCNN from Scratch or fine-tune a model using Tensorpack's training API. Observe the training with
    Tensorpack callbacks and evaluate the training progress with a validation data set after certain training intervals.

    Tensorpack provides a training API under TF1. Training runs under a TF2 installation if TF2 behavior is deactivated.

    :param path_config_yaml: path to TP config file. Check the
                             [deepdoctection.extern.tp.tpfrcnn.config.config][] for various settings.
    :param dataset_train: the dataset to use for training.
    :param path_weights: path to a checkpoint, if you want to continue training or fine-tune. Will train from scratch if
                         nothing is passed.
    :param config_overwrite: Pass a list of arguments if some configs from the .yaml file should be replaced. Use the
                             list convention, e.g. ['TRAIN.STEPS_PER_EPOCH=500', 'OUTPUT.RESULT_SCORE_THRESH=0.4']
    :param log_dir: Path to log dir. Will default to TRAIN.LOG_DIR
    :param build_train_config: dataflow build setting. Again, use list convention setting, e.g. ['max_datapoints=1000']
    :param dataset_val: the dataset to use for validation.
    :param build_val_config: same as 'build_train_config' but for validation
    :param metric_name: A metric name to choose for validation. Will use the default setting. If you want a custom
                        metric setting pass a metric explicitly.
    :param metric: A metric to choose for validation.
    :param pipeline_component_name: A pipeline component to use for validation.
    """

    assert disable_tfv2()  # TP works only in Graph mode

    build_train_dict: Dict[str, str] = {}
    if build_train_config is not None:
        build_train_dict = string_to_dict(",".join(build_train_config))
    if "split" not in build_train_dict:
        build_train_dict["split"] = "train"

    build_val_dict: Dict[str, str] = {}
    if build_val_config is not None:
        build_val_dict = string_to_dict(",".join(build_val_config))
    if "split" not in build_val_dict:
        build_val_dict["split"] = "val"

    config_overwrite = [] if config_overwrite is None else config_overwrite

    log_dir = "TRAIN.LOG_DIR=" + log_dir
    config_overwrite.append(log_dir)

    config = set_config_by_yaml(path_config_yaml)

    if config_overwrite:
        config.update_args(config_overwrite)

    categories = dataset_train.dataflow.categories.get_categories(filtered=True)
    model_frcnn_config(config, categories, False)
    model = ResNetFPNModel(config=config)

    warmup_schedule, lr_schedule, step_number = train_frcnn_config(config)

    train_dataflow = get_train_dataflow(dataset_train, config, True, **build_train_dict)
    # This is what's commonly referred to as "epochs"

    try:
        size = len(train_dataflow)
        total_passes = config.TRAIN.LR_SCHEDULE[-1] * 8 / size
        logger.info("Total passes of the training set is: %i", total_passes)

    except NotImplementedError:
        logger.info("Cannot evaluate size of dataflow and total passes")

    # Create callbacks ...

    callbacks = [
        PeriodicCallback(
            ModelSaver(max_to_keep=10, keep_checkpoint_every_n_hours=1, checkpoint_dir=config.TRAIN.LOG_DIR),
            every_k_epochs=config.TRAIN.CHECKPOINT_PERIOD,
        ),
        # linear warmup
        ScheduledHyperParamSetter("learning_rate", warmup_schedule, interp="linear", step_based=True),
        ScheduledHyperParamSetter("learning_rate", lr_schedule),
        GPUMemoryTracker(),
        HostMemoryTracker(),
        ThroughputTracker(samples_per_step=config.TRAIN.NUM_GPUS),
        EstimatedTimeLeft(median=True),
        SessionRunTimeout(60000),
        GPUUtilizationTracker(),
    ]

    if (
        config.TRAIN.EVAL_PERIOD > 0
        and dataset_val is not None
        and (metric_name is not None or metric is not None)
        and pipeline_component_name is not None
    ):
        if metric_name is not None:
            metric = metric_registry.get(metric_name)
        categories = dataset_val.dataflow.categories.get_categories(filtered=True)
        detector = TPFrcnnDetector(
            path_config_yaml,
            path_weights,
            categories,
            config_overwrite,
            True,
        )  # only a wrapper for the predictor itself. Will be replaced in Callback
        pipeline_component_cls = pipeline_component_registry.get(pipeline_component_name)
        pipeline_component = pipeline_component_cls(detector)
        assert isinstance(pipeline_component, PredictorPipelineComponent)
        category_names = list(categories.values())
        callbacks.extend(
            [
                EvalCallback(
                    dataset_val,
                    category_names,
                    dataset_val.dataflow.categories.cat_to_sub_cat,
                    metric,  # type: ignore
                    pipeline_component,
                    *model.get_inference_tensor_names(),  # type: ignore
                    **build_val_dict
                )
            ]
        )

    session_init = SmartInit(path_weights, ignore_mismatch=True)

    factor = 8.0 / config.TRAIN.NUM_GPUS

    train_cfg = TrainConfig(
        model=model,
        data=QueueInput(train_dataflow),
        callbacks=callbacks,
        steps_per_epoch=step_number,
        max_epoch=config.TRAIN.LR_SCHEDULE[-1] * factor // step_number,
        session_init=session_init,
        starting_epoch=config.TRAIN.STARTING_EPOCH,
    )

    trainer = SyncMultiGPUTrainerReplicated(config.TRAIN.NUM_GPUS, average=False)
    launch_train_with_config(train_cfg, trainer)



# -*- coding: utf-8 -*-
# File: metacfg.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Class AttrDict for maintaining configs and some functions for generating and saving AttrDict instances to .yaml files
"""

import pprint
from typing import Any, Dict, List

import yaml

from .detection_types import Pathlike


# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")
class AttrDict:
    """
    Class for storing key,values as instance with attributes and values.
    """

    _freezed = False

    # Avoid accidental creation of new hierarchies.

    def __getattr__(self, name: str) -> Any:
        """
        __getattr__
        """
        if self._freezed:
            raise AttributeError(name)
        if name.startswith("_"):
            # Do not mess with internals. Otherwise, copy/pickle will fail
            raise AttributeError(name)
        ret = AttrDict()
        setattr(self, name, ret)
        return ret

    def __setattr__(self, name: str, value: Any) -> None:
        """
        __setattr__
        """
        if self._freezed and name not in self.__dict__:
            raise AttributeError(f"Config was freezed! Unknown config: {name}")
        super().__setattr__(name, value)

    def __str__(self) -> str:
        """
        __str__
        """
        return pprint.pformat(self.to_dict(), width=100, compact=True)

    __repr__ = __str__

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a nested dict."""
        return {
            k: v.to_dict() if isinstance(v, AttrDict) else v for k, v in self.__dict__.items() if not k.startswith("_")
        }

    def from_dict(self, d: Dict[str, Any]) -> None:  # pylint: disable=C0103
        """
        Generate an instance from a dict
        """
        if isinstance(d, dict):
            self.freeze(False)
            for k, v in d.items():  # pylint: disable=C0103
                self_v = getattr(self, k)
                if isinstance(v, dict):
                    self_v.from_dict(v)
                else:
                    setattr(self, k, v)

    def update_args(self, args: List[str]) -> None:
        """
        Update from command line args.
        """
        for cfg in args:
            keys, v = cfg.split("=", maxsplit=1)  # pylint: disable=C0103
            key_list = keys.split(".")

            dic = self
            for _, k in enumerate(key_list[:-1]):
                assert k in dir(dic), f"Unknown config key: {keys}"
                dic = getattr(dic, k)
            key = key_list[-1]

            old_v = getattr(dic, key)
            if not isinstance(old_v, str):
                v = eval(v)  # pylint: disable=C0103, W0123
            setattr(dic, key, v)

    def freeze(self, freezed: bool = True) -> None:
        """
        :param freezed: freeze the instance, so that no attributes can be added or changed
        """
        self._freezed = freezed
        for v in self.__dict__.values():  # pylint: disable=C0103
            if isinstance(v, AttrDict):
                v.freeze(freezed)

    # avoid silent bugs
    def __eq__(self, _: Any) -> bool:
        raise NotImplementedError()

    def __ne__(self, _: Any) -> bool:
        raise NotImplementedError()


def set_config_by_yaml(path_yaml: Pathlike) -> AttrDict:
    """
    Use to initialize the config class for tensorpack faster rcnn

    :param path_yaml: The path to the file
    """
    config = AttrDict()
    _C = config  # pylint: disable=C0103
    _C.freeze(freezed=False)

    with open(path_yaml, "r") as file:  # pylint: disable=W1514
        _C.from_dict(yaml.load(file, Loader=yaml.Loader))

    _C.freeze()
    return config


def save_config_to_yaml(config: AttrDict, path_yaml: Pathlike) -> None:
    """
    :param config: The configuration instance as an AttrDict
    :param path_yaml: Save the config class for tensorpack faster rcnn
    :return: yaml_path: The path to save the file to
    """

    with open(path_yaml, "w") as file:  # pylint: disable=W1514
        yaml.dump(config.to_dict(), file)


def config_to_cli_str(config: AttrDict, *exclude: str) -> str:
    """
    Transform an AttrDict to a string that can be passed to a cli. Add optionally keys of the config that should not be
    added to the string.

    :param config: An `AttrDict`
    :param exclude: keys of the AttrDict
    :return: A string that can be passed to a cli
    """

    config_dict = config.to_dict()
    for key in exclude:
        config_dict.pop(key)

    output_str = ""
    for key, val in config_dict.items():
        output_str += f"--{key} {val} "

    return output_str
